"""

 snapshot_safari_importer_reprise.py

 This is a 2023 update to snapshot_safari_importer.py.  We do a bunch of things now that
 we didn't do the last time we imported Snapshot data (like updating the big taxonomy)
 file, and we skip a bunch of things now that we used to do (like generating massive
 zipfiles).  So, new year, new importer.

"""

#%% Constants and imports

import os
import glob
import json
import shutil
import random

import pandas as pd

from tqdm import tqdm
from collections import defaultdict

from megadetector.utils import path_utils

input_base = '/media/user/Elements'
output_base = os.path.expanduser('~/data/snapshot-safari-metadata')
file_list_cache_file = os.path.join(output_base,'file_list.json')

assert os.path.isdir(input_base)
os.makedirs(output_base,exist_ok=True)

# We're going to copy all the .csv files to a faster location
annotation_cache_dir = os.path.join(output_base,'csv_files')
os.makedirs(annotation_cache_dir,exist_ok=True)


#%% List files

# Do a one-time enumeration of the entire drive; this will take a long time,
# but will save a lot of hassle later.

if os.path.isfile(file_list_cache_file):
    print('Loading file list from {}'.format(file_list_cache_file))
    with open(file_list_cache_file,'r') as f:
        all_files = json.load(f)    
else:
    all_files = glob.glob(os.path.join(input_base,'**','*.*'),recursive=True)
    all_files = [fn for fn in all_files if '$RECYCLE.BIN' not in fn]
    all_files = [fn for fn in all_files if 'System Volume Information' not in fn]
    print('Enumerated {} files'.format(len(all_files)))
    with open(file_list_cache_file,'w') as f:
        json.dump(all_files,f,indent=1)
    print('Wrote file list to {}'.format(file_list_cache_file))


#%% Create derived lists

# Takes about 60 seconds

all_files_relative = [os.path.relpath(fn,input_base) for fn in all_files]
all_files_relative = [fn.replace('\\','/') for fn in all_files_relative]
all_files_relative_set = set(all_files_relative)

# CSV files are one of:
#
# _report_lila.csv (species/count/etc. for each capture)
# _report_lila_image_inventory.csv (maps captures to images)
# _report_lila_overview.csv (distribution of species)
csv_files = [fn for fn in all_files_relative if fn.endswith('.csv')]

all_image_files = path_utils.find_image_strings(all_files_relative)

print('Found a total of {} files, {} of which are images'.format(
    len(all_files_relative),len(all_image_files)))


#%% Copy all csv files to the annotation cache folder

# fn = csv_files[0]
for fn in csv_files:
    target_file = os.path.join(annotation_cache_dir,os.path.basename(fn))
    source_file = os.path.join(input_base,fn)
    shutil.copyfile(source_file,target_file)    

def read_cached_csv_file(fn):
    """
    Later cells will ask to read a .csv file from the original hard drive;
    read from the annotation cache instead.
    """
    
    cached_csv_file = os.path.join(annotation_cache_dir,os.path.basename(fn))
    df = pd.read_csv(cached_csv_file)
    return df


#%% List project folders

# There are two formats for project folder names:
#
# APN
# Snapshot Cameo/DEB
project_code_to_project_folder = {}

folders = os.listdir(input_base)
folders = [fn for fn in folders if (not fn.startswith('$') and \
                                    not 'System Volume' in fn)]

for fn in folders:
    if len(fn) == 3:
        assert fn not in project_code_to_project_folder
        project_code_to_project_folder[fn] = fn
    else:
        assert 'Snapshot' in fn
        subfolders = os.listdir('/'.join([input_base,fn]))
        for subfn in subfolders:
            assert len(subfn) == 3
            assert subfn not in project_code_to_project_folder
            project_code_to_project_folder[subfn] = '/'.join([fn,subfn])

project_folder_to_project_code = {v: k for k, v in project_code_to_project_folder.items()}
project_codes = sorted(list(project_code_to_project_folder.keys()))
project_folders = sorted(list(project_code_to_project_folder.values()))

def file_to_project_folder(fn):
    """
    For a given filename relative to the drive root, return the corresponding
    project folder (also relative to the drive root).
    """
    
    tokens = fn.split('/')
    if len(tokens[0]) == 3:
        project_folder = tokens[0]
    else:
        assert 'Snapshot' in tokens[0]
        project_folder = '/'.join(tokens[0:2])
    assert project_folder in project_folders
    return project_folder


def file_to_project_code(fn):
    """
    For a given filename relative to the drive root, return the corresponding
    three-letter project code (e.g. "CDB").
    """    
    
    return project_folder_to_project_code[file_to_project_folder(fn)]

assert file_to_project_folder(
    'APN/APN_S2/DW/DW_R5/APN_S2_DW_R5_IMAG0003.JPG') == 'APN'
assert file_to_project_folder(
    'Snapshot South Africa/BLO/BLO_S1/B05/B05_R1/BLO_S1_B05_R1_IMAG0003.JPG') == \
    'Snapshot South Africa/BLO'
assert file_to_project_code(
    'Snapshot South Africa/BLO/BLO_S1/B05/B05_R1/BLO_S1_B05_R1_IMAG0003.JPG') == \
    'BLO'
    
    
#%% Map report and inventory files to codes

# Maps a three-letter project code to a list of per-season _report_lila.csv files
#
# E.g.:
#
# 'DHP': ['Snapshot South Africa/DHP/LILA_Reports/DHP_S1_report_lila.csv',
# 'Snapshot South Africa/DHP/LILA_Reports/DHP_S2_report_lila.csv',
# 'Snapshot South Africa/DHP/LILA_Reports/DHP_S3_report_lila.csv']
#
project_code_to_report_files = defaultdict(list)

# fn = csv_files[0]
for fn in csv_files:
    if 'report_lila.csv' not in fn:
        continue
    project_code = project_folder_to_project_code[file_to_project_folder(fn)]
    project_code_to_report_files[project_code].append(fn)

project_codes_with_no_reports = set()

for project_code in project_code_to_project_folder.keys():
    if project_code not in project_code_to_report_files:
        project_codes_with_no_reports.add(project_code)
        print('Warning: no report files available for {}'.format(project_code))


#%% Make sure that every report has a corresponding inventory file

all_report_files = [item for sublist in project_code_to_report_files.values() \
                    for item in sublist]

for fn in all_report_files:
    inventory_file = fn.replace('.csv','_image_inventory.csv')
    assert inventory_file in csv_files
    
    
#%% Count species based on overview and report files

# The overview and report files should produce the same counts; we'll verify this
# in the next cell.

species_to_count_overview = defaultdict(int)
species_to_count_report = defaultdict(int)

for report_file in all_report_files:
        
    overview_file = report_file.replace('.csv','_overview.csv')    
    
    df = read_cached_csv_file(overview_file)
    
    for i_row,row in df.iterrows():
        
        if row['question'] == 'question__species':
            
            assert isinstance(row['answer'],str)
            assert isinstance(row['count'],int)
            species = row['answer']
            
            if len(species) < 3:
                assert species == '0' or species == '1'
                
            species_to_count_overview[species] += row['count']
    
    # ...for each capture in the overview file
    
    df = read_cached_csv_file(report_file)
    
    for i_row,row in df.iterrows():
        
        species = row['question__species']
        assert isinstance(species,str)
        
        # Ignore results from the blank/non-blank workflow
        if len(species) < 3:
            assert species == '0' or species == '1'                
        species_to_count_report[species] += 1
                    
    # ...for each capture in the report file
    
# ...for each report file
    

#%% Print counts

species_to_count_overview_sorted = \
    {k: v for k, v in sorted(species_to_count_overview.items(), 
                             key=lambda item: item[1], reverse=True)}
species_to_count_report_sorted = \
    {k: v for k, v in sorted(species_to_count_report.items(), 
                             key=lambda item: item[1], reverse=True)}

string_count = 0
non_blank_count = 0

for species in species_to_count_overview_sorted.keys():        
    
    # The overview and report files should produce the same counts
    assert species_to_count_overview_sorted[species] == \
        species_to_count_report[species]    
    count = species_to_count_overview_sorted[species]
    if species not in ('0','1'):
        string_count += count
        if species != 'blank':
            non_blank_count += count
            
    print('{}{}'.format(species.ljust(25),count))

n_images = len(all_files)
n_sequences = sum(species_to_count_overview_sorted.values())

print('\n{} total images\n{} total sequences'.format(n_images,n_sequences))

print('\nString count: {}'.format(string_count))
print('Non-blank count: {}'.format(non_blank_count))


#%% Make sure that capture IDs in the reports/inventory files match

# ...and confirm that (almost) all the images in the inventory tables are 
# present on disk.

all_relative_paths_in_inventory = set()
files_missing_on_disk = []

for report_file in all_report_files:
        
    project_base = file_to_project_folder(report_file)                
    inventory_file = report_file.replace('.csv','_image_inventory.csv')    
    
    inventory_df = read_cached_csv_file(inventory_file)
    report_df = read_cached_csv_file(report_file)
    
    capture_ids_in_report = set()
    for i_row,row in report_df.iterrows():
        capture_ids_in_report.add(row['capture_id'])
    
    capture_ids_in_inventory = set()
    for i_row,row in inventory_df.iterrows():
        
        capture_ids_in_inventory.add(row['capture_id'])
        image_path_relative = project_base + '/' + row['image_path_rel']
        
        # assert image_path_relative in all_files_relative_set
        if image_path_relative not in all_files_relative_set:
            
            # Make sure this isn't just a case issue
            assert image_path_relative.replace('.JPG','.jpg') \
                not in all_files_relative_set
            assert image_path_relative.replace('.jpg','.JPG') \
                not in all_files_relative_set
            files_missing_on_disk.append(image_path_relative)
            
        assert image_path_relative not in all_relative_paths_in_inventory
        all_relative_paths_in_inventory.add(image_path_relative)
                
    # Make sure the set of capture IDs appearing in this report is
    # the same as the set of capture IDs appearing in the corresponding
    # inventory file.
    assert capture_ids_in_report == capture_ids_in_inventory

# ...for each report file

print('\n{} missing files (of {})'.format(
    len(files_missing_on_disk),len(all_relative_paths_in_inventory)))

    
#%% For all the files we have on disk, see which are and aren't in the inventory files

# There aren't any capital-P .PNG files, but if I don't include .PNG
# in this list, I'll look at this in a year and wonder whether I forgot
# to include it.
image_extensions = set(['.JPG','.jpg','.PNG','.png'])

images_not_in_inventory = []
n_images_in_inventoried_projects = 0

# fn = all_files_relative[0]
for fn in tqdm(all_files_relative):

    if os.path.splitext(fn)[1] not in image_extensions:
        continue
    project_code = file_to_project_code(fn)
    if project_code in project_codes_with_no_reports:
        # print('Skipping project {}'.format(project_code))
        continue
    n_images_in_inventoried_projects += 1
    if fn not in all_relative_paths_in_inventory:
        images_not_in_inventory.append(fn)

print('\n{} images on disk are not in inventory (of {} in eligible projects)'.format(
    len(images_not_in_inventory),n_images_in_inventoried_projects))


#%% Map captures to images, and vice-versa

capture_id_to_images = defaultdict(list)
image_to_capture_id = {}

# report_file = all_report_files[0]
for report_file in tqdm(all_report_files):
        
    inventory_file = report_file.replace('.csv','_image_inventory.csv')    
    inventory_df = read_cached_csv_file(inventory_file)
    
    project_folder = file_to_project_folder(inventory_file)
    
    # row = inventory_df.iloc[0]
    for i_row,row in inventory_df.iterrows():
    
        capture_id = row['capture_id']
        image_file_relative = os.path.join(project_folder,row['image_path_rel'])
        capture_id_to_images[capture_id].append(image_file_relative)
        assert image_file_relative not in image_to_capture_id
        image_to_capture_id[image_file_relative] = capture_id
        
    # ...for each row (one image per row)
        
# ...for each report file
    

#%% Map captures to species (just species for now, we'll go back and get other metadata later)

capture_id_to_species = defaultdict(list)

for project_code in tqdm(project_codes):
    
    report_files = project_code_to_report_files[project_code]
    
    for report_file in report_files:
        
        report_df = read_cached_csv_file(report_file)

        for i_row,row in report_df.iterrows():
        
            capture_id = row['capture_id']
            species = row['question__species']
            capture_id_to_species[capture_id].append(species)

        # ...for each row
        
    # ...for each report file in this project
    
# ...for each project


#%% Take a look at the annotations "0" and "1"

captures_0 = []
captures_1 = []
captures_1_alone = []
captures_1_with_species = []

for capture_id in tqdm(capture_id_to_species):
    
    species_this_capture_id = capture_id_to_species[capture_id]
    
    # Multiple rows may be present for a capture, but they should be unique
    assert len(species_this_capture_id) == len(set(species_this_capture_id))
    
    if '0' in species_this_capture_id:
        captures_0.append(capture_id)
        # '0' should always appear alone
        assert len(species_this_capture_id) == 1

    if '1' in species_this_capture_id:
        captures_1.append(capture_id)
        assert '0' not in species_this_capture_id
        # '1' should never appear alone
        # assert len(species_this_capture_id) > 1
        if len(species_this_capture_id) == 1:
            captures_1_alone.append(capture_id)
        else:
            captures_1_with_species.append(capture_id)

# ...for each capture ID

print('')
print('Number of captures with "0" as the species (always appears alone): {}'.format(len(captures_0)))
print('Number of captures with "1" as the species: {}'.format(len(captures_1)))
print('Number of captures with "1" as the species, with no other species: {}'.format(
    len(captures_1_alone)))
print('Number of captures with "1" as the species, with other species: {}'.format(
    len(captures_1_with_species)))


#%% Sample some of those captures with mysterious "0" and "1" annotations

random.seed(0)
n_to_sample = 500
captures_0_samples = random.sample(captures_0,n_to_sample)
captures_1_samples = random.sample(captures_1,n_to_sample)

capture_0_sample_output_folder = os.path.join(output_base,'capture_0_samples')
capture_1_sample_output_folder = os.path.join(output_base,'capture_1_samples')
os.makedirs(capture_0_sample_output_folder,exist_ok=True)
os.makedirs(capture_1_sample_output_folder,exist_ok=True)

def copy_sampled_captures(sampled_captures,sample_capture_output_folder):

    for capture_id in tqdm(sampled_captures):    
        images_this_capture = capture_id_to_images[capture_id]
        for fn in images_this_capture:            
            # assert fn in all_files_relative_set
            if fn not in all_files_relative_set:
                print('Warning: missing file {}'.format(fn))
                continue
            source_image = os.path.join(input_base,fn)
            target_image = os.path.join(sample_capture_output_folder,os.path.basename(fn))
            shutil.copyfile(source_image,target_image)
        # ....for each image
    # ...for each capture
        
copy_sampled_captures(captures_0_samples,capture_0_sample_output_folder)    
copy_sampled_captures(captures_1_samples,capture_1_sample_output_folder)


#%% Find images that MD thinks contain people

md_results_folder = os.path.expanduser(
    '~/postprocessing/snapshot-safari/snapshot-safari-2023-04-21-v5a.0.0/json_subsets')
md_results_files = os.listdir(md_results_folder)

md_human_detection_threshold = 0.2
md_vehicle_detection_threshold = 0.2

# We'll make sure this is actually correct for all the files we load
md_human_category = '2'
md_vehicle_category = '3'

md_human_images = set()
md_vehicle_images = set()

# project_code = project_codes[0]
for project_code in project_codes:
    
    print('Finding human images for {}'.format(project_code))
    
    project_folder = project_code_to_project_folder[project_code]
    
    md_results_file = [fn for fn in md_results_files if project_code in fn]
    assert len(md_results_file) == 1
    md_results_file = os.path.join(md_results_folder,md_results_file[0])
    
    with open(md_results_file,'r') as f:
        md_results = json.load(f)
    assert md_results['detection_categories'][md_human_category] == 'person'
    assert md_results['detection_categories'][md_vehicle_category] == 'vehicle'
    
    # im = md_results['images'][0]
    for im in tqdm(md_results['images']):
        
        if 'detections' not in im:
            continue
        
        # MD results files are each relative to their own projects, we want
        # filenames to be relative to the base of the drive
        fn = os.path.join(project_folder,im['file'])
        for det in im['detections']:
            if det['category'] == md_human_category and \
                det['conf'] >= md_human_detection_threshold:
                    md_human_images.add(fn)
            if det['category'] == md_vehicle_category and \
                det['conf'] >= md_vehicle_detection_threshold:
                    md_vehicle_images.add(fn)
                    
        # ...for each detection
        
    # ...for each image
                    
# ...for each project

print('MD found {} human images, {} vehicle images'.format(
    len(md_human_images),len(md_vehicle_images)))

md_human_or_vehicle_images = \
    set(md_human_images).union(set(md_vehicle_images))
    
# next(iter(md_human_or_vehicle_images))


#%% Find images where the ground truth says humans or vehicles are present

human_species_id = 'human'
vehicle_species_id = 'humanvehicle'

gt_human_capture_ids = set()
gt_vehicle_capture_ids = set()

for capture_id in capture_id_to_species:
    
    species_this_capture_id = capture_id_to_species[capture_id]
    
    for species in species_this_capture_id:
        if species == human_species_id:
            gt_human_capture_ids.add(capture_id)
        elif species == vehicle_species_id:
            gt_vehicle_capture_ids.add(capture_id)
    
# ...for each capture ID

gt_human_images = []
gt_vehicle_images = []

for capture_id in gt_human_capture_ids:
    images_this_capture_id = capture_id_to_images[capture_id]
    gt_human_images.extend(images_this_capture_id)
for capture_id in gt_vehicle_capture_ids:
    images_this_capture_id = capture_id_to_images[capture_id]
    gt_vehicle_images.extend(images_this_capture_id)    
    
print('Ground truth includes {} human images ({} captures), {} vehicle images ({} captures)'.format(
    len(gt_human_images),len(gt_human_capture_ids),
    len(gt_vehicle_images),len(gt_vehicle_capture_ids)))

ground_truth_human_or_vehicle_images = \
    set(gt_human_images).union(set(gt_vehicle_images))
    
# next(iter(ground_truth_human_or_vehicle_images))


#%% Find mismatches

gt_missing_human_images = []
gt_missing_vehicle_images = []

for fn in md_human_images:
    if fn not in ground_truth_human_or_vehicle_images:
        gt_missing_human_images.append(fn)

for fn in md_vehicle_images:
    if fn not in ground_truth_human_or_vehicle_images:
        gt_missing_vehicle_images.append(fn)
        
print('Of {} images where MD found a human, {} are not in the ground truth'.format(
    len(md_human_images),len(gt_missing_human_images)))

print('Of {} images where MD found a vehicle, {} are not in the ground truth'.format(
    len(md_vehicle_images),len(gt_missing_vehicle_images)))


#%% Sample mismatches

random.seed(0)
n_to_sample = 1000
sampled_human_mismatches = random.sample(gt_missing_human_images,n_to_sample)
sampled_vehicle_mismatches = random.sample(gt_missing_vehicle_images,n_to_sample)

human_mismatch_output_folder = os.path.join(output_base,'mismatches_human')
vehicle_mismatch_output_folder = os.path.join(output_base,'mismatches_vehicle')
os.makedirs(human_mismatch_output_folder,exist_ok=True)
os.makedirs(vehicle_mismatch_output_folder,exist_ok=True)

def copy_sampled_images(sampled_images,sampled_images_output_folder):

    for fn in tqdm(sampled_images): 
        if fn not in all_files_relative_set:
            print('Warning: missing file {}'.format(fn))
            continue
        source_image = os.path.join(input_base,fn)
        target_image = os.path.join(sampled_images_output_folder,os.path.basename(fn))
        shutil.copyfile(source_image,target_image)
    
copy_sampled_images(sampled_human_mismatches,human_mismatch_output_folder)
copy_sampled_images(sampled_vehicle_mismatches,vehicle_mismatch_output_folder)


#%% See what's up with some of the mismatches

filename_base_to_filename = {}

from megadetector.utils.path_utils import is_image_file

# fn = all_files_relative[0]
for fn in tqdm(all_files_relative):

    if not is_image_file(fn):
        continue
    if 'Indiv_Recognition' in fn:
        continue
    bn = os.path.basename(fn)
    assert bn not in filename_base_to_filename
    filename_base_to_filename[bn] = fn


if False:

    bn = 'TSW_S2_KA02_R3_IMAG0002.JPG'
    fn = filename_base_to_filename[bn]
    capture_id = image_to_capture_id[fn]
    species = capture_id_to_species[capture_id]
    
    
#%% Look at the distribution of labels for the mismatched images    

gt_missing_images = set(gt_missing_human_images).union(set(gt_missing_vehicle_images))

missing_image_species_to_count = defaultdict(int)

for fn in gt_missing_images:
    if fn not in image_to_capture_id:
        continue
    capture_id = image_to_capture_id[fn]
    species = capture_id_to_species[capture_id]
    for s in species:
        missing_image_species_to_count[s] += 1



## Camera trap real-time API configuration

REDIS_HOST = 'localhost'

REDIS_PORT = 6379

# Full path to the temporary folder for image storage, only meaningful 
# within the Docker container
TEMP_FOLDER = '/app/temp'

REDIS_QUEUE_NAME = 'camera-trap-queue'

# Upper limit on total content length (all images and parameters)
MAX_CONTENT_LENGTH_IN_MB = 5 * 8  # 5MB per image * number of images allowed

MAX_IMAGES_ACCEPTED = 8

IMAGE_CONTENT_TYPES = ['image/png', 'application/octet-stream', 'image/jpeg']

DETECTOR_MODEL_PATH = '/app/animal_detection_api/model/md_v5a.0.0.pt'

DETECTOR_MODEL_VERSION = 'v5a.0.0'

# Minimum confidence threshold for detections
DEFAULT_CONFIDENCE_THRESHOLD = 0.01

# Minimum confidence threshold for showing a bounding box on the output image
DEFAULT_RENDERING_CONFIDENCE_THRESHOLD = 0.2

API_PREFIX = '/v1/camera-trap/sync'

API_KEYS_FILE = 'allowed_keys.txt'

# Use this when testing without Docker
DETECTOR_MODEL_PATH_DEBUG = 'model/md_v5a.0.0.pt'



"""

 idfg_iwildcam_lila_prep.py

 Adding class labels (from the private test .csv) to the iWildCam 2019 IDFG 
 test set, in preparation for release on LILA.

 This version works with the public iWildCam release images.

"""

#%% ############ Take one, from iWildCam .json files ############ 

#%% Imports and constants

import uuid
import json
import os
from tqdm import tqdm

base_folder = r'h:\iWildCam_2019_IDFG'
input_json = os.path.join(base_folder,'iWildCam_2019_IDFG_info.json')
input_csv = os.path.join(base_folder,'IDFG_eval_public_v_private.csv')
output_json = os.path.join(base_folder,'idaho_camera_traps.json')

assert os.path.isfile(input_json)
assert os.path.isfile(input_csv)


#%% Read input files

with open(input_json,'r') as f:
    input_data = json.load(f)

with open(input_csv,'r') as f:
    private_csv_lines = f.readlines()

private_csv_lines = [s.strip() for s in private_csv_lines]
    
# Remove the header line    
assert private_csv_lines[0] == 'Id,Category,Usage'
private_csv_lines = private_csv_lines[1:]

print('Read {} annotations for {} images'.format(len(private_csv_lines),len(input_data['images'])))

assert len(private_csv_lines) == len(input_data['images'])
n_images = len(input_data['images'])


#%% Parse annotations

image_id_to_category_ids = {}
for line in tqdm(private_csv_lines):
    
    # Lines look like:
    #
    # b005e5b2-2c0b-11e9-bcad-06f1011196c4,1,Private
    
    tokens = line.split(',')
    assert len(tokens) == 3
    assert tokens[2] in ['Private','Public']
    image_id_to_category_ids[tokens[0]] = int(tokens[1])
    
assert len(image_id_to_category_ids) == n_images


#%% Minor cleanup re: images

for im in tqdm(input_data['images']):
    image_id = im['id']
    im['file_name'] = im['file_name'].replace('iWildCam_IDFG_images/','')
    assert isinstance(im['location'],int)
    im['location'] = str(im['location'])
    

#%% Create annotations

annotations = []

for image_id in tqdm(image_id_to_category_ids):
    category_id = image_id_to_category_ids[image_id]
    ann = {}
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = image_id
    ann['category_id'] = category_id
    annotations.append(ann)
    

#%% Prepare info

info = input_data['info']
info['contributor'] = 'Images acquired by the Idaho Department of Fish and Game, dataset curated by Sara Beery'
info['description'] = 'Idaho Camera traps'
info['version'] = '2021.07.19'


#%% Minor adjustments to categories

input_categories = input_data['categories']

category_id_to_name = {cat['id']:cat['name'] for cat in input_categories}
category_name_to_id = {cat['name']:cat['id'] for cat in input_categories}
assert category_id_to_name[0] == 'empty'

category_names_to_counts = {}
for category in input_categories:
    category_names_to_counts[category['name']] = 0
    
for ann in annotations:
    category_id = ann['category_id']
    category_name = category_id_to_name[category_id]
    category_names_to_counts[category_name] = category_names_to_counts[category_name] + 1
    
categories = []

for category_name in category_names_to_counts:
    count = category_names_to_counts[category_name]    
    
    # Remove unused categories
    if count == 0:
        continue
    
    category_id = category_name_to_id[category_name]
    
    # Name adjustments
    if category_name == 'prongs':
        category_name = 'pronghorn'
    
    categories.append({'id':category_id,'name':category_name})
    
    
#%% Create output

output_data = {}
output_data['images'] = input_data['images']
output_data['annotations'] = annotations
output_data['categories'] = categories
output_data['info'] = info


#%% Write output

with open(output_json,'w') as f:
    json.dump(output_data,f,indent=2)
    

#%% Validate .json file

from megadetector.data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = os.path.join(base_folder,'images'); assert os.path.isdir(options.baseDir)
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False

_, _, _ = integrity_check_json_db.integrity_check_json_db(output_json, options)


#%% Preview labels

from megadetector.visualization import visualize_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 100
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.include_filename_links = True

# viz_options.classes_to_exclude = ['test']
html_output_file, _ = visualize_db.visualize_db(db_path=output_json,
                                                         output_dir=os.path.join(
                                                         base_folder,'preview'),
                                                         image_base_dir=os.path.join(base_folder,'images'),
                                                         options=viz_options)
os.startfile(html_output_file)


#%% ############ Take two, from pre-iWildCam .json files created from IDFG .csv files ############ 

#%% Imports and constants

import json
import os

base_folder = r'h:\idaho-camera-traps'
input_json_sl = os.path.join(base_folder,'iWildCam_IDFG.json')
input_json = os.path.join(base_folder,'iWildCam_IDFG_ml.json')
output_json = os.path.join(base_folder,'idaho_camera_traps.json')
remote_image_base_dir = r'z:\idfg'

assert os.path.isfile(input_json)


#%% One-time line break addition

if not os.path.isfile(input_json):
    
    sl_json = input_json_sl
    ml_json = input_json
    
    with open(sl_json,'r') as f:
        d = json.load(f)
    with open(ml_json,'w') as f:
        json.dump(d,f,indent=2)
        
        
#%% Read input files

with open(input_json,'r') as f:
    input_data = json.load(f)

print('Read {} annotations for {} images'.format(len(input_data['annotations']),len(input_data['images'])))

n_images = len(input_data['images'])


#%% Prepare info

info = {}
info['contributor'] = 'Images acquired by the Idaho Department of Fish and Game, dataset curated by Sara Beery'
info['description'] = 'Idaho Camera traps'
info['version'] = '2021.07.19'


#%% Minor adjustments to categories

input_categories = input_data['categories']
output_categories = []

for c in input_categories:
    category_name = c['name']
    category_id = c['id']
    if category_name == 'prong':
        category_name = 'pronghorn'
    category_name = category_name.lower()
    output_categories.append({'name':category_name,'id':category_id})


#%% Minor adjustments to annotations

for ann in input_data['annotations']:
    ann['id'] = str(ann['id'])


#%% Create output

output_data = {}
output_data['images'] = input_data['images']
output_data['annotations'] = input_data['annotations']
output_data['categories'] = output_categories
output_data['info'] = info


#%% Write output

with open(output_json,'w') as f:
    json.dump(output_data,f,indent=2)
    

#%% Validate .json file

from megadetector.data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = remote_image_base_dir
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False

_, _, _ = integrity_check_json_db.integrity_check_json_db(output_json, options)


#%% Preview labels

from megadetector.visualization import visualize_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 100
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.include_filename_links = True

# viz_options.classes_to_exclude = ['test']
html_output_file, _ = visualize_db.visualize_db(db_path=output_json,
                                                         output_dir=os.path.join(
                                                         base_folder,'preview'),
                                                         image_base_dir=remote_image_base_dir,
                                                         options=viz_options)
os.startfile(html_output_file)



"""

remap_coco_categories.py

Given a COCO-formatted dataset, remap the categories to a new mapping.

"""

#%% Imports and constants

import os
import json

from copy import deepcopy


#%% Main function

def remap_coco_categories(input_data,
                          output_category_name_to_id,
                          input_category_name_to_output_category_name,
                          output_file=None):
    """
    Given a COCO-formatted dataset, remap the categories to a new categories mapping, optionally
    writing the results to a new file.
    
    output_category_name_to_id is a dict mapping strings to ints.
    
    input_category_name_to_output_category_name is a dict mapping strings to strings.
    
    [input_data] can be a COCO-formatted dict or a filename.  If it's a dict, it will be copied,
    not modified in place.
    """
    
    if isinstance(input_data,str):
        assert os.path.isfile(input_data), "Can't find file {}".format(input_data)
        with open(input_data,'r') as f:
            input_data = json.load(f)
        assert isinstance(input_data,dict), 'Illegal COCO input data'
    else:
        assert isinstance(input_data,dict), 'Illegal COCO input data'
        input_data = deepcopy(input_data)
    
    # It's safe to modify in-place now
    output_data = input_data
    
    # Read input name --> ID mapping
    input_category_name_to_input_category_id = {}
    for c in input_data['categories']:
        input_category_name_to_input_category_id[c['name']] = c['id']
    
    # Map input IDs --> output IDs
    input_category_id_to_output_category_id = {}
    for input_name in input_category_name_to_output_category_name.keys():
        output_name = input_category_name_to_output_category_name[input_name]
        assert output_name in output_category_name_to_id, \
            'No output ID for {} --> {}'.format(input_name,output_name)
        input_id = input_category_name_to_input_category_id[input_name]
        output_id = output_category_name_to_id[output_name]
        input_category_id_to_output_category_id[input_id] = output_id
        
    # Map annotations
    for ann in output_data['annotations']:
        assert ann['category_id'] in input_category_id_to_output_category_id, \
            'Unrecognized category ID {}'.format(ann['category_id'])
        ann['category_id'] = input_category_id_to_output_category_id[ann['category_id']]
        
    # Update the category list
    output_categories = []
    for output_name in output_category_name_to_id:
        category = {'name':output_name,'id':output_category_name_to_id[output_name]}
        output_categories.append(category)
    output_data['categories'] = output_categories
        
    if output_file is not None:
        with open(output_file,'w') as f:
            json.dump(output_data,f,indent=1)
            
    return input_data


#%% Command-line driver

# TODO



"""

 auckland_doc_test_to_json.py

 Convert Auckland DOC data set to COCO camera traps format.  This was
 for a testing data set where a .csv file was provided with class
 information.

"""

#%% Constants and imports

import json
import os
import uuid
import pandas as pd
import datetime
import ntpath
import re
import numpy as np

from tqdm import tqdm

from megadetector.visualization import visualize_db
from megadetector.data_management.databases import integrity_check_json_db
from megadetector.utils.path_utils import find_images

input_base_dir = r'e:\auckland-test\2_Testing'

input_metadata_file = r'G:\auckland-doc\Maukahuka - Auckland Island - Cat camera data Master April 2019 - DOC-5924483.xlsx'

# Filenames will be stored in the output .json relative to this base dir
output_base_dir = r'g:\auckland-doc'
output_json_filename = os.path.join(output_base_dir, 'auckland-doc-test.json')

assert os.path.isdir(input_base_dir)
os.makedirs(output_base_dir,exist_ok=True)

output_encoding = 'utf-8'
read_image_sizes = True

info = {}
info['year'] = 2020
info['version'] = '1.0'
info['description'] = 'Auckaland DOC Camera Traps (test)'
info['contributor'] = 'Auckland DOC'
info['date_created'] = str(datetime.date.today())


#%% Enumerate files

print('Enumerating files from {}'.format(input_base_dir))
absolute_image_paths = find_images(input_base_dir, recursive=True)
print('Enumerated {} images'.format(len(absolute_image_paths)))

relative_image_paths = []
for fn in absolute_image_paths:
    relative_image_paths.append(os.path.relpath(fn,input_base_dir).replace('\\','/'))

relative_image_paths_set = set(relative_image_paths)

assert len(relative_image_paths_set) == len(relative_image_paths)


#%% Create unique identifier for each image

# The ground truth doesn't have full paths in it; create unique identifiers for each image 
# based on the camera name and filename.
# 
# We store file identifiers as cameraname_filename.
file_identifier_to_relative_paths = {}
camera_names = set()

# relative_path = relative_image_paths[0]
for relative_path in relative_image_paths:
    
    # Example relative paths
    #
    # Summer_Trial_2019/A1_1_42_SD114_20190210/AucklandIsland_A1_1_42_SD114_20190210_01300001.jpg
    # Winter_Trial_2019/Installation/10_F4/10_F4_tmp_201908210001.JPG
    fn = ntpath.basename(relative_path)    
    
    # Find the camera name
    tokens = relative_path.split('/')
    
    if tokens[1] == 'Installation' or 'Rebait' in tokens[1]:
        camera_name = tokens[2]
    
    else:
        # E..g. "A1_1_42_SD114_20190210" in the above example
        camera_token = tokens[1]
        camera_name = None
        m = re.search('^(.+)_SD',camera_token)
        if m:
            camera_name = m.group(1)
        else:
            # For camera tokens like C1_5_D_190207
            m = re.search('^(.+_.+_.+)',camera_token)
            camera_name = m.group(1)
    
    assert camera_name
    camera_names.add(camera_name)
    
    file_identifier = camera_name + '_' + fn
    if file_identifier not in file_identifier_to_relative_paths:
        file_identifier_to_relative_paths[file_identifier] = [relative_path]
    else:
        file_identifier_to_relative_paths[file_identifier].append(relative_path)
    
print('Found {} unique camera names'.format(len(camera_names)))
      
      
#%% Load input data

input_metadata = pd.read_excel(input_metadata_file)

print('Read {} columns and {} rows from metadata file'.format(len(input_metadata.columns),
      len(input_metadata)))

# The spreadsheet has a space after "Camera"
input_metadata = input_metadata.rename(columns={'Camera ':'Camera'})
    

#%% Assemble dictionaries

image_id_to_image = {}
category_name_to_category = {}
annotations = []

# Force the empty category to be ID 0
empty_category = {}
empty_category['name'] = 'empty'
empty_category['id'] = 0
category_name_to_category['empty'] = empty_category

rows_not_found_in_folder = []
rows_ambiguous = []
rows_no_filename = []
rows_no_annotation = []

image_id_to_rows = {}

next_id = 1

category_names = ['cat','mouse','unknown','human','pig','sealion','penguin','dog','openadjusted']

# array([nan, 'Blackbird', 'Bellbird', 'Tomtit', 'Song thrush', 'Pippit',
#       'Pippet', '?', 'Dunnock', 'Song thursh', 'Kakariki', 'Tui', ' ',
#       'Silvereye', 'NZ Pipit', 'Blackbird and Dunnock', 'Unknown',
#       'Pipit', 'Songthrush'], dtype=object)

def bird_name_to_category_name(bird_name):
    bird_name = bird_name.lower().strip().replace(' ','_').replace('song_thursh','song_thrush')
    bird_name = bird_name.replace('pippet','pipt').replace('pippit','pipit').replace('nz_pipit','pipit')
    if bird_name == '?' or bird_name == '' or bird_name == 'unknown':
        category_name = 'unknown_bird'
    else:
        category_name = bird_name
    return category_name
    
bird_names = input_metadata.Bird_ID.unique()
for bird_name in bird_names:
    if isinstance(bird_name,float):
        continue
    category_name = bird_name_to_category_name(bird_name)
    if category_name not in category_names:
        category_names.append(category_name)
        
for category_name in category_names:
    cat = {}
    cat['name'] = category_name
    cat['id'] = next_id
    next_id = next_id +1
    category_name_to_category[category_name] = cat
    
def create_annotation(image_id,category_name,count):
    assert isinstance(image_id,str)
    assert isinstance(category_name,str)
    assert isinstance(count,int) or isinstance(count,float)
    if isinstance(count,float):
        count = int(count)
    ann = {}    
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = image_id
    category = category_name_to_category[category_name]
    category_id = category['id']
    ann['category_id'] = category_id
    ann['count'] = count
    return ann
    
# i_row = 0; row = input_metadata.iloc[i_row]
for i_row,row in tqdm(input_metadata.iterrows(),total=len(input_metadata)):

    # E.g.: AucklandIsland_A1_1_42_SD114_20190210_01300009.jpg
    filename = row['File']
    if isinstance(filename,float):
        rows_no_filename.append(i_row)
        continue
    
    camera_name = row['Camera']
    file_identifier = camera_name + '_' + filename
    if not file_identifier in file_identifier_to_relative_paths:
        rows_not_found_in_folder.append(i_row)
        continue        

    relative_paths_this_file_id = file_identifier_to_relative_paths[file_identifier]
    
    if len(relative_paths_this_file_id) == 1:
        relative_path = relative_paths_this_file_id[0]
    else:
        
        # We have multiple files matching this identifier, can we uniquely resolve this
        # to one of those files based on the camera ID?
        matches = [s for s in relative_paths_this_file_id if camera_name in s]
        assert len(matches) > 0
        if len(matches) > 1:
            rows_ambiguous.append(i_row)
            continue
        relative_path = matches[0]

    assert filename.endswith('.jpg') or filename.endswith('.JPG')
    image_id = filename.lower().replace('.jpg','')

    if image_id in image_id_to_rows:
        image_id_to_rows[image_id].append(i_row)
        continue
        
    image_id_to_rows[image_id] = [i_row]
        
    im = {}
    im['id'] = image_id 
    im['file_name'] = relative_path
    im['datetime'] = str(row['Date and time'])
    im['camera'] = row['Camera']
    im['sd_card'] = row['SD_Card']
    im['sd_change'] = row['SD_Change']
    im['comments'] = row['Comments']
    
    image_id_to_image[im['id']] = im
    
    # create_annotation(image_id,category_name,count)
    
    # 'SD_Change', 'Cat', 'Mouse', 'Bird', 'Bird_ID', 'False_trig', 'Unknown',
    # 'Human', 'Collared_cat', 'Cat_ID', 'Pig', 'Sea_lion', 'Open_adjusted',
    # 'Penguin', 'Dog', 'Comments', 'Unnamed: 22']
    
    # Each of these categories is handled a little differently...
    
    annotations_this_image = []
    if (not np.isnan(row['Cat'])):
        assert np.isnan(row['Collared_cat'] )
        annotations_this_image.append(create_annotation(im['id'],'cat',row['Cat']))
    
    if (not np.isnan(row['Collared_cat'])):
        assert np.isnan(row['Cat'] )
        annotations_this_image.append(create_annotation(im['id'],'cat',row['Collared_cat']))
    
    if (not np.isnan(row['Bird'])):
        if isinstance(row['Bird_ID'],str):
            category_name = bird_name_to_category_name(row['Bird_ID'])
        else:
            assert np.isnan(row['Bird_ID'])            
            category_name = 'unknown_bird'
        annotations_this_image.append(create_annotation(im['id'],category_name,row['Bird']))
    
    if (not np.isnan(row['False_trig'])):
        annotations_this_image.append(create_annotation(im['id'],'empty',-1))
    
    # These are straightforward
    for s in ['Mouse','Unknown','Pig','Human','Sea_lion','Penguin','Dog','Open_adjusted']:
        if isinstance(row[s],float) or isinstance(row[s],int):
            if not np.isnan(row[s]):
                 annotations_this_image.append(create_annotation(im['id'],s.lower().replace('_',''),row[s]))
        elif isinstance(row[s],str):
            print('File {}, label {}, value {}'.format(filename,s,row[s]))
        else:
            raise ValueError('Error handling count value {}'.format(row[s]))
        
    if len(annotations_this_image) > 1:
        print('Multiple annotations for filename {}'.format(filename))
       
    if len(annotations_this_image) == 0:
        rows_no_annotation.append(i_row)
        
    annotations.extend(annotations_this_image)

# ...for each image


#%% Summarize errors

print('Of {} rows:\n'.format(len(input_metadata)))

print('{} images not found in folder'.format(len(rows_not_found_in_folder)))
print('{} images ambiguously mapped'.format(len(rows_ambiguous)))
print('{} images no filename'.format(len(rows_no_filename)))
print('{} images no annotation'.format(len(rows_no_annotation)))
print('{} images handled successfully, {} total annotations'.format(len(image_id_to_image),len(annotations)))
    

#%% Write output .json

images = list(image_id_to_image.values())
categories = list(category_name_to_category.values())

data = {}
data['info'] = info
data['images'] = images
data['annotations'] = annotations
data['categories'] = categories

json.dump(data, open(output_json_filename, 'w'), indent=1)
print('Finished writing json to {}'.format(output_json_filename))


#%% Validate .json file

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = input_base_dir
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False

sortedCategories, data, _ = integrity_check_json_db.integrity_check_json_db(output_json_filename, options)


#%% Preview labels

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 2000
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.classes_to_exclude = ['empty']
html_output_file, image_db = visualize_db.visualize_db(db_path=output_json_filename,
                                                         output_dir=os.path.join(
                                                         output_base_dir, 'preview'),
                                                         image_base_dir=input_base_dir,
                                                         options=viz_options)
os.startfile(html_output_file)


#%% Precision-recall analysis

from megadetector.postprocessing.postprocess_batch_results import PostProcessingOptions
from megadetector.postprocessing.postprocess_batch_results import process_batch_results

api_output_file = r'g:\auckland-doc\auckland-doc_20200801\combined_api_outputs\auckland-doc_202008012020.08.01_reformatMaukahuka_Auckland_Island2_TestingSummer_Trial_2019_detections.filtered_rde_0.60_0.85_5_0.05.json'
postprocessing_output_folder = r'G:\auckland-doc\auckland-doc_20200801\postprocessing'
image_base = r'E:\auckland-test\2_Testing'
ground_truth_json_file = output_json_filename

output_base = os.path.join(postprocessing_output_folder,'pr_analysis')
os.makedirs(output_base,exist_ok=True)

options = PostProcessingOptions()
options.unlabeled_classes.append('openadjusted')
options.image_base_dir = image_base
options.parallelize_rendering = True
options.include_almost_detections = True
options.num_images_to_sample = 2500
options.confidence_threshold = 0.75
options.almost_detection_confidence_threshold = 0.7
options.ground_truth_json_file = ground_truth_json_file
options.allow_missing_images = True
options.ground_truth_filename_replacements = {} 
options.api_output_filename_replacements = {'2020.08.01_reformat\\Maukahuka_Auckland_Island\\2_Testing\\':''}
options.api_output_file = api_output_file
options.output_dir = output_base
ppresults = process_batch_results(options)
os.startfile(ppresults.output_html_file)




"""

run_classifier.py

Run a species classifier.

This script is the classifier counterpart to detection/run_tf_detector_batch.py.
This script takes as input:
1) a detections JSON file, usually the output of run_tf_detector_batch.py or the
    output of the Batch API in the "Batch processing API output format"
2) a path to a directory containing crops of bounding boxes from the detections
    JSON file
3) a path to a PyTorch TorchScript compiled model file
4) (if the model is EfficientNet) an image size

By default, this script overwrites the detections JSON file, adding in
classification results. To output a new JSON file, use the --output argument.

"""

#%% Imports

from __future__ import annotations

import argparse
import json
import os

from tqdm import tqdm
from typing import Any
from collections.abc import Callable, Sequence

import pandas as pd
import PIL
import torch
import torch.utils
import torchvision as tv
from torchvision.datasets.folder import default_loader

from megadetector.classification import train_classifier


#%% Example usage

"""
    python run_classifier.py \
        detections.json \
        /path/to/crops \
        /path/to/model.pt \
        --image-size 224
"""


#%% Classes

class SimpleDataset(torch.utils.data.Dataset):
    """
    Very simple dataset.
    """

    def __init__(self, img_files: Sequence[str],
                 images_dir: str | None = None,
                 transform: Callable[[PIL.Image.Image], Any] | None = None):
        """Creates a SimpleDataset."""
        self.img_files = img_files
        self.images_dir = images_dir
        self.transform = transform

    def __getitem__(self, index: int) -> tuple[Any, str]:
        """
        Returns: tuple, (img, img_file)
        """
        img_file = self.img_files[index]
        if self.images_dir is not None:
            img_path = os.path.join(self.images_dir, img_file)
        else:
            img_path = img_file
        img = default_loader(img_path)
        if self.transform is not None:
            img = self.transform(img)
        return img, img_file

    def __len__(self) -> int:
        return len(self.img_files)


#%% Support functions

def create_loader(cropped_images_dir: str,
                  detections_json_path: str | None,
                  img_size: int,
                  batch_size: int,
                  num_workers: int
                  ) -> torch.utils.data.DataLoader:
    """
    Creates a DataLoader.

    Args:
        cropped_images_dir: str, path to image crops
        detections_json_path: optional str, path to detections JSON
        img_size: int, resizes smallest side of image to img_size,
            then center-crops to (img_size, img_size)
        batch_size: int, batch size in dataloader
        num_workers: int, # of workers in dataloader
    """
    
    crop_files = []

    if detections_json_path is None:
        # recursively find all files in cropped_images_dir
        for subdir, _, files in os.walk(cropped_images_dir):
            for file_name in files:
                rel_dir = os.path.relpath(subdir, cropped_images_dir)
                rel_file = os.path.join(rel_dir, file_name)
                crop_files.append(rel_file)

    else:
        # only find crops of images from detections JSON
        print('Loading detections JSON')
        with open(detections_json_path, 'r') as f:
            js = json.load(f)
        detections = {img['file']: img for img in js['images']}
        detector_version = js['info']['detector']

        for img_file, info_dict in tqdm(detections.items()):
            if 'detections' not in info_dict or info_dict['detections'] is None:
                continue
            for i in range(len(info_dict['detections'])):
                crop_filename = img_file + f'___crop{i:02d}_{detector_version}.jpg'
                crop_path = os.path.join(cropped_images_dir, crop_filename)
                if os.path.exists(crop_path):
                    crop_files.append(crop_filename)

    transform = tv.transforms.Compose([
        # resizes smaller edge to img_size
        tv.transforms.Resize(img_size, interpolation=PIL.Image.BICUBIC),
        tv.transforms.CenterCrop(img_size),
        tv.transforms.ToTensor(),
        tv.transforms.Normalize(mean=train_classifier.MEANS,
                                std=train_classifier.STDS, inplace=True)
    ])

    dataset = SimpleDataset(img_files=crop_files, images_dir=cropped_images_dir,
                            transform=transform)
    assert len(dataset) > 0
    loader = torch.utils.data.DataLoader(
        dataset, batch_size=batch_size, num_workers=num_workers,
        pin_memory=True)
    return loader


#%% Main function

def main(model_path: str,
         cropped_images_dir: str,
         output_csv_path: str,
         detections_json_path: str | None,
         classifier_categories_json_path: str | None,
         img_size: int,
         batch_size: int,
         num_workers: int,
         device_id: int | None = None) -> None:
    
    # Evaluating with accimage is much faster than Pillow or Pillow-SIMD, but accimage
    # is Linux-only.
    try:
        import accimage # noqa
        tv.set_image_backend('accimage')
    except:
        print('Warning: could not start accimage backend (ignore this if you\'re not using Linux)')

    # create dataset
    print('Creating data loader')
    loader = create_loader(
        cropped_images_dir, detections_json_path=detections_json_path,
        img_size=img_size, batch_size=batch_size, num_workers=num_workers)

    label_names = None
    if classifier_categories_json_path is not None:
        with open(classifier_categories_json_path, 'r') as f:
            categories = json.load(f)
        label_names = [categories[str(i)] for i in range(len(categories))]

    # create model
    print('Loading saved model')
    model = torch.jit.load(model_path)
    model, device = train_classifier.prep_device(model, device_id=device_id)

    test_epoch(model, loader, device=device, label_names=label_names,
               output_csv_path=output_csv_path)


def test_epoch(model: torch.nn.Module,
               loader: torch.utils.data.DataLoader,
               device: torch.device,
               label_names: Sequence[str] | None,
               output_csv_path: str) -> None:
    """
    Runs for 1 epoch.

    Writes results to the output CSV in batches.

    Args:
        model: torch.nn.Module
        loader: torch.utils.data.DataLoader
        device: torch.device
        label_names: optional list of str, label names
        output_csv_path: str
    """
    
    # set dropout and BN layers to eval mode
    model.eval()

    header = True
    mode = 'w'  # new file on first write

    with torch.no_grad():
        for inputs, img_files in tqdm(loader):
            inputs = inputs.to(device, non_blocking=True)
            outputs = model(inputs)
            probs = torch.nn.functional.softmax(outputs, dim=1).cpu().numpy()

            if label_names is None:
                label_names = [str(i) for i in range(probs.shape[1])]

            df = pd.DataFrame(data=probs, columns=label_names,
                              index=pd.Index(img_files, name='path'))
            df.to_csv(output_csv_path, index=True, header=header, mode=mode)

            if header:
                header = False
                mode = 'a'


#%% Command-line driver

def _parse_args() -> argparse.Namespace:

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Run classifier.')
    parser.add_argument(
        'model',
        help='path to TorchScript compiled model')
    parser.add_argument(
        'crops_dir',
        help='path to directory containing cropped images')
    parser.add_argument(
        'output',
        help='path to save CSV file with classifier results (can use .csv.gz '
             'extension for compression)')
    parser.add_argument(
        '-d', '--detections-json',
        help='path to detections JSON file, used to filter paths within '
             'crops_dir')
    parser.add_argument(
        '-c', '--classifier-categories',
        help='path to JSON file for classifier categories. If not given, '
             'classes are numbered "0", "1", "2", ...')
    parser.add_argument(
        '--image-size', type=int, default=224,
        help='size of input image to model, usually 224px, but may be larger '
             'especially for EfficientNet models')
    parser.add_argument(
        '--batch-size', type=int, default=1,
        help='batch size for evaluating model')
    parser.add_argument(
        '--device', type=int, default=None,
        help='preferred CUDA device')
    parser.add_argument(
        '--num-workers', type=int, default=8,
        help='# of workers for data loading')
    return parser.parse_args()


if __name__ == '__main__':
    
    args = _parse_args()
    main(model_path=args.model,
         cropped_images_dir=args.crops_dir,
         output_csv_path=args.output,
         detections_json_path=args.detections_json,
         classifier_categories_json_path=args.classifier_categories,
         img_size=args.image_size,
         batch_size=args.batch_size,
         num_workers=args.num_workers,
         device_id=args.device)



"""

 animl_results_to_md_results.py

 Convert a .csv file produced by the Animl package:

 https://github.com/conservationtechlab/animl-py

 ...to a MD results file suitable for import into Timelapse.

 Columns are expected to be:

 file
 category (MD category identifies: 1==animal, 2==person, 3==vehicle)
 detection_conf
 bbox1,bbox2,bbox3,bbox4
 class
 classification_conf

"""

#%% Imports and constants

import pandas as pd
import json

# It's a little icky to hard-code this here rather than importing from elsewhere
# in the MD repo, but it seemed silly to take a dependency on lots of MD code
# just for this, so, hard-coding.
detection_category_id_to_name = {'1':'animal','2':'person','3':'vehicle'}


#%% Main function

def animl_results_to_md_results(input_file,output_file=None):
    """
    Converts the .csv file [input_file] to the MD-formatted .json file [output_file].
    
    If [output_file] is None, '.json' will be appended to the input file.
    """
    
    if output_file is None:
        output_file = input_file + '.json'

    df = pd.read_csv(input_file)
    
    expected_columns = ('file','category','detection_conf',
                        'bbox1','bbox2','bbox3','bbox4','class','classification_conf')
    
    for s in expected_columns:
        assert s in df.columns,\
            'Expected column {} not found'.format(s)
            
    classification_category_name_to_id = {}    
    filename_to_results = {}
    
    # i_row = 0; row = df.iloc[i_row]
    for i_row,row in df.iterrows():
            
        # Is this the first detection we've seen for this file?            
        if row['file'] in filename_to_results:
            im = filename_to_results[row['file']]
        else:
            im = {}
            im['detections'] = []
            im['file'] = row['file']
            filename_to_results[im['file']] = im
            
        assert isinstance(row['category'],int),'Invalid category identifier in row {}'.format(im['file'])
        detection_category_id = str(row['category'])
        assert detection_category_id in detection_category_id_to_name,\
            'Unrecognized detection category ID {}'.format(detection_category_id)
            
        detection = {}
        detection['category'] = detection_category_id
        detection['conf'] = row['detection_conf']
        bbox = [row['bbox1'],row['bbox2'],row['bbox3'],row['bbox4']]
        detection['bbox'] = bbox
        classification_category_name = row['class']
        
        # Have we seen this classification category before?
        if classification_category_name in classification_category_name_to_id:
            classification_category_id = \
                classification_category_name_to_id[classification_category_name]
        else:
            classification_category_id = str(len(classification_category_name_to_id))
            classification_category_name_to_id[classification_category_name] = \
                classification_category_id
        
        classifications = [[classification_category_id,row['classification_conf']]]
        detection['classifications'] = classifications
        
        im['detections'].append(detection)
        
    # ...for each row        
        
    info = {}
    info['format_version'] = '1.3'
    info['detector'] = 'Animl'
    info['classifier'] = 'Animl'
    
    results = {}
    results['info'] = info
    results['detection_categories'] = detection_category_id_to_name
    results['classification_categories'] = \
        {v: k for k, v in classification_category_name_to_id.items()}
    results['images'] = list(filename_to_results.values())
    
    with open(output_file,'w') as f:
        json.dump(results,f,indent=1)
        
# ...animl_results_to_md_results(...)
        

#%% Interactive driver

if False:

    pass

    #%%
    
    input_file = r"G:\temp\animl-runs\animl-runs\Coati_v2\manifest.csv"
    output_file = None
    animl_results_to_md_results(input_file,output_file)
    

#%% Command-line driver

import sys,argparse

def main():

    parser = argparse.ArgumentParser(
        description='Convert an Animl-formatted .csv results file to MD-formatted .json results file')
    
    parser.add_argument(
        'input_file',
        type=str,
        help='input .csv file')
    
    parser.add_argument(
        '--output_file',
        type=str,
        default=None,
        help='output .json file (defaults to input file appended with ".json")')
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    animl_results_to_md_results(args.input_file,args.output_file)
    
if __name__ == '__main__':
    main()
    



"""

path_utils.py

Miscellaneous useful utils for path manipulation, i.e. things that could *almost*
be in os.path, but aren't.

"""

#%% Imports and constants

import glob
import ntpath
import os
import sys
import platform
import string
import json
import shutil
import unicodedata
import zipfile
import tarfile
import webbrowser
import subprocess
import re

from zipfile import ZipFile
from datetime import datetime
from multiprocessing.pool import Pool, ThreadPool
from functools import partial
from shutil import which
from tqdm import tqdm

# Should all be lower-case
IMG_EXTENSIONS = ('.jpg', '.jpeg', '.gif', '.png', '.tif', '.tiff', '.bmp')

VALID_FILENAME_CHARS = f"~-_.() {string.ascii_letters}{string.digits}"
SEPARATOR_CHARS = r":\/"
VALID_PATH_CHARS = VALID_FILENAME_CHARS + SEPARATOR_CHARS
CHAR_LIMIT = 255


#%% General path functions

def recursive_file_list(base_dir, 
                        convert_slashes=True, 
                        return_relative_paths=False, 
                        sort_files=True,
                        recursive=True):
    r"""
    Enumerates files (not directories) in [base_dir], optionally converting
    backslahes to slashes
    
    Args:
        base_dir (str): folder to enumerate
        convert_slashes (bool, optional): force forward slashes; if this is False, will use
            the native path separator
        return_relative_paths (bool, optional): return paths that are relative to [base_dir],
            rather than absolute paths
        sort_files (bool, optional): force files to be sorted, otherwise uses the sorting
            provided by os.walk()
        recursive (bool, optional): enumerate recursively
        
    Returns:
        list: list of filenames
    """
    
    assert os.path.isdir(base_dir), '{} is not a folder'.format(base_dir)
    
    all_files = []

    if recursive:
        for root, _, filenames in os.walk(base_dir):
            for filename in filenames:
                full_path = os.path.join(root, filename)
                all_files.append(full_path)
    else:
        all_files_relative = os.listdir(base_dir)
        all_files = [os.path.join(base_dir,fn) for fn in all_files_relative]
        all_files = [fn for fn in all_files if os.path.isfile(fn)]
        
    if return_relative_paths:
        all_files = [os.path.relpath(fn,base_dir) for fn in all_files]

    if convert_slashes:
        all_files = [fn.replace('\\', '/') for fn in all_files]
    
    if sort_files:
        all_files = sorted(all_files)
        
    return all_files


def file_list(base_dir, convert_slashes=True, return_relative_paths=False, sort_files=True, 
              recursive=False):
    """
    Trivial wrapper for recursive_file_list, which was a poor function name choice at the time, 
    since it doesn't really make sense to have a "recursive" option in a function called 
    "recursive_file_list".
    
    Args:
        base_dir (str): folder to enumerate
        convert_slashes (bool, optional): force forward slashes; if this is False, will use
            the native path separator
        return_relative_paths (bool, optional): return paths that are relative to [base_dir],
            rather than absolute paths
        sort_files (bool, optional): force files to be sorted, otherwise uses the sorting
            provided by os.walk()
        recursive (bool, optional): enumerate recursively
        
    Returns:
        list: list of filenames    
    """
    
    return recursive_file_list(base_dir,convert_slashes,return_relative_paths,sort_files,
                               recursive=recursive)


def fileparts(path):
    r"""
    Breaks down a path into the directory path, filename, and extension.

    Note that the '.' lives with the extension, and separators are removed.

    Examples:
        
    .. code-block:: none

        >>> fileparts('file')    
        ('', 'file', '')
        >>> fileparts(r'c:/dir/file.jpg')
        ('c:/dir', 'file', '.jpg')
        >>> fileparts('/dir/subdir/file.jpg')
        ('/dir/subdir', 'file', '.jpg')        

    Args:
        path (str): path name to separate into parts
    Returns:
        tuple: tuple containing (p,n,e):        
            - p: str, directory path
            - n: str, filename without extension
            - e: str, extension including the '.'
    """
    
    # ntpath seems to do the right thing for both Windows and Unix paths
    p = ntpath.dirname(path)
    basename = ntpath.basename(path)
    n, e = ntpath.splitext(basename)
    return p, n, e


def insert_before_extension(filename, s=None, separator='.'):
    """
    Insert string [s] before the extension in [filename], separated with [separator].

    If [s] is empty, generates a date/timestamp. If [filename] has no extension,
    appends [s].

    Examples:
        
    .. code-block:: none
    
        >>> insert_before_extension('/dir/subdir/file.ext', 'insert')
        '/dir/subdir/file.insert.ext'
        >>> insert_before_extension('/dir/subdir/file', 'insert')
        '/dir/subdir/file.insert'
        >>> insert_before_extension('/dir/subdir/file')
        '/dir/subdir/file.2020.07.20.10.54.38'
        
    Args:
        filename (str): filename to manipulate
        s (str, optional): string to insert before the extension in [filename], or
            None to insert a datestamp
        separator (str, optional): separator to place between the filename base
            and the inserted string
            
    Returns:
        str: modified string
    """
    
    assert len(filename) > 0
    if s is None or len(s) == 0:
        s = datetime.now().strftime('%Y.%m.%d.%H.%M.%S')
    name, ext = os.path.splitext(filename)
    return f'{name}{separator}{s}{ext}'


def split_path(path):
    r"""
    Splits [path] into all its constituent file/folder tokens.

    Examples:
        
    .. code-block:: none
    
        >>> split_path(r'c:\dir\subdir\file.txt')
        ['c:\\', 'dir', 'subdir', 'file.txt']
        >>> split_path('/dir/subdir/file.jpg')
        ['/', 'dir', 'subdir', 'file.jpg']
        >>> split_path('c:\\')
        ['c:\\']
        >>> split_path('/')
        ['/']
        
    Args:
        path (str): path to split into tokens
    
    Returns:
        list: list of path tokens
    """
    
    parts = []
    while True:
        # ntpath seems to do the right thing for both Windows and Unix paths
        head, tail = ntpath.split(path)
        if head == '' or head == path:
            break
        parts.append(tail)
        path = head
    parts.append(head or tail)
    return parts[::-1] # reverse


def path_is_abs(p):
    """
    Determines whether [p] is an absolute path.  An absolute path is defined as
    one that starts with slash, backslash, or a letter followed by a colon.
    
    Args:
        p (str): path to evaluate
        
    Returns:
        bool: True if [p] is an absolute path, else False
    """
    
    return (len(p) > 1) and (p[0] == '/' or p[1] == ':' or p[0] == '\\')


def top_level_folder(p):
    r"""
    Gets the top-level folder from the path *p*.
    
    On UNIX, this is straightforward:
        
    /blah/foo 
    
    ...returns '/blah'
    
    On Windows, we define this as the top-level folder that isn't the drive, so:
        
    c:\blah\foo
    
    ...returns 'c:\blah'.
    
    Args:
        p (str): filename to evaluate
        
    Returns:
        str: the top-level folder in [p], see above for details on how this is defined
    """
    
    if p == '':
        return ''
    
    # Path('/blah').parts is ('/','blah')
    parts = split_path(p)
    
    if len(parts) == 1:
        return parts[0]

    # Handle paths like:
    #
    # /, \, /stuff, c:, c:\stuff
    drive = os.path.splitdrive(p)[0]
    if parts[0] == drive or parts[0] == drive + '/' or parts[0] == drive + '\\' or parts[0] in ['\\', '/']:
        return os.path.join(parts[0], parts[1])
    else:
        return parts[0]

# ...top_level_folder()


#%% Test driver for top_level_folder

if False:  

    #%%
    
    p = 'blah/foo/bar'; s = top_level_folder(p); print(s); assert s == 'blah'
    p = '/blah/foo/bar'; s = top_level_folder(p); print(s); assert s == '/blah'
    p = 'bar'; s = top_level_folder(p); print(s); assert s == 'bar'
    p = ''; s = top_level_folder(p); print(s); assert s == ''
    p = 'c:\\'; s = top_level_folder(p); print(s); assert s == 'c:\\'
    p = r'c:\blah'; s = top_level_folder(p); print(s); assert s == 'c:\\blah'
    p = r'c:\foo'; s = top_level_folder(p); print(s); assert s == 'c:\\foo'
    p = r'c:/foo'; s = top_level_folder(p); print(s); assert s == 'c:/foo'
    p = r'c:\foo/bar'; s = top_level_folder(p); print(s); assert s == 'c:\\foo'
        
    #%%

def safe_create_link(link_exists,link_new):
    """
    Creates a symlink at [link_new] pointing to [link_exists].
    
    If [link_new] already exists, make sure it's a link (not a file),
    and if it has a different target than [link_exists], removes and re-creates
    it.
    
    Errors if [link_new] already exists but it's not a link.
    
    Args:
        link_exists (str): the source of the (possibly-new) symlink
        link_new (str): the target of the (possibly-new) symlink
    """
    
    if os.path.exists(link_new) or os.path.islink(link_new):
        assert os.path.islink(link_new)
        if not os.readlink(link_new) == link_exists:
            os.remove(link_new)
            os.symlink(link_exists,link_new)
    else:
        os.symlink(link_exists,link_new)
        

#%% Image-related path functions

def is_image_file(s, img_extensions=IMG_EXTENSIONS):
    """
    Checks a file's extension against a hard-coded set of image file
    extensions.  Uses case-insensitive comparison.
    
    Does not check whether the file exists, only determines whether the filename
    implies it's an image file.
    
    Args:
        s (str): filename to evaluate for image-ness
        img_extensions (list, optional): list of known image file extensions
        
    Returns:
        bool: True if [s] appears to be an image file, else False
    """
    
    ext = os.path.splitext(s)[1]
    return ext.lower() in img_extensions


def find_image_strings(strings):
    """
    Given a list of strings that are potentially image file names, looks for
    strings that actually look like image file names (based on extension).
    
    Args:
        strings (list): list of filenames to check for image-ness
        
    Returns:
        list: the subset of [strings] that appear to be image filenames
    """
    
    return [s for s in strings if is_image_file(s)]


def find_images(dirname, 
                recursive=False, 
                return_relative_paths=False, 
                convert_slashes=True):
    """
    Finds all files in a directory that look like image file names. Returns
    absolute paths unless return_relative_paths is set.  Uses the OS-native
    path separator unless convert_slashes is set, in which case will always
    use '/'.
    
    Args:
        dirname (str): the folder to search for images
        recursive (bool, optional): whether to search recursively
        return_relative_paths (str, optional): return paths that are relative
            to [dirname], rather than absolute paths
        convert_slashes (bool, optional): force forward slashes in return values

    Returns:
        list: list of image filenames found in [dirname]
    """
    
    assert os.path.isdir(dirname), '{} is not a folder'.format(dirname)
    
    if recursive:
        strings = glob.glob(os.path.join(dirname, '**', '*.*'), recursive=True)
    else:
        strings = glob.glob(os.path.join(dirname, '*.*'))
    
    image_files = find_image_strings(strings)
    
    if return_relative_paths:
        image_files = [os.path.relpath(fn,dirname) for fn in image_files]
    
    image_files = sorted(image_files)
    
    if convert_slashes:
        image_files = [fn.replace('\\', '/') for fn in image_files]
        
    return image_files


#%% Filename cleaning functions

def clean_filename(filename, 
                   allow_list=VALID_FILENAME_CHARS,
                   char_limit=CHAR_LIMIT,
                   force_lower= False):
    r"""
    Removes non-ASCII and other invalid filename characters (on any
    reasonable OS) from a filename, then optionally trims to a maximum length.

    Does not allow :\/ by default, use clean_path if you want to preserve those.

    Adapted from
    https://gist.github.com/wassname/1393c4a57cfcbf03641dbc31886123b8
    
    Args:
        filename (str): filename to clean
        allow_list (str, optional): string containing all allowable filename characters
        char_limit (int, optional): maximum allowable filename length, if None will skip this
            step
        force_lower (bool, optional): convert the resulting filename to lowercase
    
    returns:
        str: cleaned version of [filename]            
    """
    
    # keep only valid ascii chars
    cleaned_filename = (unicodedata.normalize('NFKD', filename)
                        .encode('ASCII', 'ignore').decode())

    # keep only allow-listed chars
    cleaned_filename = ''.join([c for c in cleaned_filename if c in allow_list])
    if char_limit is not None:
        cleaned_filename = cleaned_filename[:char_limit]
    if force_lower:
        cleaned_filename = cleaned_filename.lower()
    return cleaned_filename


def clean_path(pathname, 
               allow_list=VALID_PATH_CHARS,
               char_limit=CHAR_LIMIT,
               force_lower=False):
    """
    Removes non-ASCII and other invalid path characters (on any reasonable
    OS) from a path, then optionally trims to a maximum length.
    
    Args:
        pathname (str): path name to clean
        allow_list (str, optional): string containing all allowable filename characters
        char_limit (int, optional): maximum allowable filename length, if None will skip this
            step
        force_lower (bool, optional): convert the resulting filename to lowercase
    
    returns:
        str: cleaned version of [filename]            
    """
    
    return clean_filename(pathname, allow_list=allow_list, 
                          char_limit=char_limit, force_lower=force_lower)


def flatten_path(pathname,separator_chars=SEPARATOR_CHARS,separator_char_replacement='~'):
    r"""
    Removes non-ASCII and other invalid path characters (on any reasonable
    OS) from a path, then trims to a maximum length. Replaces all valid
    separators with [separator_char_replacement.]
    
    Args:
        pathname (str): path name to flatten
        separator_chars (str, optional): string containing all known path separators
        separator_char_replacement (str, optional): string to insert in place of 
            path separators.
            
    Returns:
        str: flattened version of [pathname]
    """
    
    s = clean_path(pathname)
    for c in separator_chars:
        s = s.replace(c, separator_char_replacement)
    return s


def is_executable(filename):    
    """
    Checks whether [filename] is on the system path and marked as executable.
    
    Args:
        filename (str): filename to check for executable status
    
    Returns:
        bool: True if [filename] is on the system path and marked as executable, otherwise False
    """
    
    # https://stackoverflow.com/questions/11210104/check-if-a-program-exists-from-a-python-script

    return which(filename) is not None


#%% Platform-independent way to open files in their associated application

def environment_is_wsl():
    """
    Determines whether we're running in WSL.
    
    Returns:
        True if we're running in WSL.        
    """
    
    if sys.platform not in ('linux','posix'):
        return False
    platform_string = ' '.join(platform.uname()).lower()
    return 'microsoft' in platform_string and 'wsl' in platform_string
    

def wsl_path_to_windows_path(filename):
    r"""
    Converts a WSL path to a Windows path, or returns None if that's not possible.  E.g.
    converts:
        
    /mnt/e/a/b/c
    
    ...to:
        
    e:\a\b\c
    
    Args:
        filename (str): filename to convert
    
    Returns:
        str: Windows equivalent to the WSL path [filename]
    """
    
    result = subprocess.run(['wslpath', '-w', filename], text=True, capture_output=True)
    if result.returncode != 0:
        print('Could not convert path {} from WSL to Windows'.format(filename))
        return None
    return result.stdout.strip()
    
    
def open_file(filename, attempt_to_open_in_wsl_host=False, browser_name=None):
    """
    Opens [filename] in the default OS file handler for this file type.
    
    If browser_name is not None, uses the webbrowser module to open the filename
    in the specified browser; see https://docs.python.org/3/library/webbrowser.html
    for supported browsers.  Falls back to the default file handler if webbrowser.open()
    fails.  In this case, attempt_to_open_in_wsl_host is ignored unless webbrowser.open() fails.
    
    If browser_name is 'default', uses the system default.  This is different from the 
    parameter to webbrowser.get(), where None implies the system default.
    
    Args:
        filename (str): file to open
        attempt_to_open_in_wsl_host: if this is True, and we're in WSL, attempts to open
            [filename] in the Windows host environment
        browser_name: see above
    """
    
    if browser_name is not None:
        if browser_name == 'chrome':
            browser_name = 'google-chrome'
        elif browser_name == 'default':
            browser_name = None
        try:
            result = webbrowser.get(using=browser_name).open(filename)
        except Exception:
            result = False
        if result:
            return
        
    if sys.platform == 'win32':
        
        os.startfile(filename)

    elif sys.platform == 'darwin':
      
        opener = 'open'
        subprocess.call([opener, filename])
            
    elif attempt_to_open_in_wsl_host and environment_is_wsl():
        
        windows_path = wsl_path_to_windows_path(filename)
        
        # Fall back to xdg-open
        if windows_path is None:
            subprocess.call(['xdg-open', filename])
            
        if os.path.isdir(filename):            
            subprocess.run(["explorer.exe", windows_path])
        else:
            os.system("cmd.exe /C start %s" % (re.escape(windows_path)))    
        
    else:
        
        opener = 'xdg-open'        
        subprocess.call([opener, filename])
        

#%% File list functions

def write_list_to_file(output_file,strings):
    """
    Writes a list of strings to either a JSON file or text file,
    depending on extension of the given file name.
    
    Args:
        output_file (str): file to write
        strings (list): list of strings to write to [output_file]
    """
    
    with open(output_file, 'w') as f:
        if output_file.endswith('.json'):
            json.dump(strings, f, indent=1)
        else:
            f.write('\n'.join(strings))


def read_list_from_file(filename):
    """
    Reads a json-formatted list of strings from a file.
    
    Args:
        filename (str): .json filename to read
    
    Returns:
        list: list of strings read from [filename]
    """
    
    assert filename.endswith('.json')
    with open(filename, 'r') as f:
        file_list = json.load(f)
    assert isinstance(file_list, list)
    for s in file_list:
        assert isinstance(s, str)
    return file_list


def _copy_file(input_output_tuple,overwrite=True,verbose=False):
    """
    Internal function for copying files from within parallel_copy_files.
    """
    
    assert len(input_output_tuple) == 2
    source_fn = input_output_tuple[0]
    target_fn = input_output_tuple[1]
    if (not overwrite) and (os.path.isfile(target_fn)):
        if verbose:
            print('Skipping existing file {}'.format(target_fn))
        return
    os.makedirs(os.path.dirname(target_fn),exist_ok=True)
    shutil.copyfile(source_fn,target_fn)
    

def parallel_copy_files(input_file_to_output_file, max_workers=16, 
                        use_threads=True, overwrite=False, verbose=False):
    """
    Copies files from source to target according to the dict input_file_to_output_file.
    
    Args:
        input_file_to_output_file (dict): dictionary mapping source files to the target files
            to which they should be copied
        max_workers (int, optional): number of concurrent workers; set to <=1 to disable parallelism
        use_threads (bool, optional): whether to use threads (True) or processes (False) for
            parallel copying; ignored if max_workers <= 1
        overwrite (bool, optional): whether to overwrite existing destination files
        verbose (bool, optional): enable additionald debug output
    """

    n_workers = min(max_workers,len(input_file_to_output_file))
    
    # Package the dictionary as a set of 2-tuples
    input_output_tuples = []
    for input_fn in input_file_to_output_file:
        input_output_tuples.append((input_fn,input_file_to_output_file[input_fn]))

    if use_threads:
        pool = ThreadPool(n_workers)
    else:
        pool = Pool(n_workers)

    with tqdm(total=len(input_output_tuples)) as pbar:
        for i,_ in enumerate(pool.imap_unordered(partial(_copy_file,overwrite=overwrite,verbose=verbose),
                                                 input_output_tuples)):
            pbar.update()

# ...def parallel_copy_files(...)


def get_file_sizes(base_dir, convert_slashes=True):
    """
    Gets sizes recursively for all files in base_dir, returning a dict mapping
    relative filenames to size.
    
    TODO: merge the functionality here with parallel_get_file_sizes, which uses slightly
    different semantics.
    
    Args:
        base_dir (str): folder within which we want all file sizes
        convert_slashes (bool, optional): force forward slashes in return strings,
            otherwise uses the native path separator
            
    Returns:
        dict: dictionary mapping filenames to file sizes in bytes
    """
    
    relative_filenames = recursive_file_list(base_dir, convert_slashes=convert_slashes, 
                                             return_relative_paths=True)
    
    fn_to_size = {}
    for fn_relative in tqdm(relative_filenames):
        fn_abs = os.path.join(base_dir,fn_relative)
        fn_to_size[fn_relative] = os.path.getsize(fn_abs)
                   
    return fn_to_size
        

def _get_file_size(filename,verbose=False):
    """
    Internal function for safely getting the size of a file.  Returns a (filename,size)
    tuple, where size is None if there is an error.
    """
    
    try:
        size = os.path.getsize(filename)
    except Exception as e:
        if verbose:
            print('Error reading file size for {}: {}'.format(filename,str(e)))
        size = None
    return (filename,size)

    
def parallel_get_file_sizes(filenames, 
                            max_workers=16, 
                            use_threads=True, 
                            verbose=False,
                            recursive=True, 
                            convert_slashes=True,
                            return_relative_paths=False):
    """
    Returns a dictionary mapping every file in [filenames] to the corresponding file size,
    or None for errors.  If [filenames] is a folder, will enumerate the folder (optionally recursively).
        
    Args:
        filenames (list or str): list of filenames for which we should read sizes, or a folder
            within which we should read all file sizes recursively
        max_workers (int, optional): number of concurrent workers; set to <=1 to disable parallelism
        use_threads (bool, optional): whether to use threads (True) or processes (False) for
            parallel copying; ignored if max_workers <= 1
        verbose (bool, optional): enable additionald debug output
        recursive (bool, optional): enumerate recursively, only relevant if [filenames] is a folder.
        convert_slashes (bool, optional): convert backslashes to forward slashes
        return_relative_paths (bool, optional): return relative paths; only relevant if [filenames]
            is a folder.
        
    Returns:
        dict: dictionary mapping filenames to file sizes in bytes
    """

    n_workers = min(max_workers,len(filenames))
    
    folder_name = None
    
    if verbose:
        print('Enumerating files')
        
    if isinstance(filenames,str) and os.path.isdir(filenames):
        
        folder_name = filenames
        
        # Enumerate absolute paths here, we'll convert to relative later if requested
        filenames = recursive_file_list(filenames,recursive=recursive,return_relative_paths=False)

    if verbose:
        print('Creating worker pool')
    
    if use_threads:
        pool_string = 'thread'
        pool = ThreadPool(n_workers)
    else:
        pool_string = 'process'
        pool = Pool(n_workers)

    if verbose:
        print('Created a {} pool of {} workers'.format(
            pool_string,n_workers))
        
    # This returns (filename,size) tuples
    get_size_results = list(tqdm(pool.imap(
        partial(_get_file_size,verbose=verbose),filenames), total=len(filenames)))
    
    to_return = {}
    for r in get_size_results:
        fn = r[0]
        if return_relative_paths and (folder_name is not None):
            fn = os.path.relpath(fn,folder_name)
        if convert_slashes:
            fn = fn.replace('\\','/')
        size = r[1]
        to_return[fn] = size

    return to_return


#%% Zip functions

def zip_file(input_fn, output_fn=None, overwrite=False, verbose=False, compresslevel=9):
    """
    Zips a single file.
    
    Args:
        input_fn (str): file to zip
        output_fn (str, optional): target zipfile; if this is None, we'll use
            [input_fn].zip
        overwrite (bool, optional): whether to overwrite an existing target file
        verbose (bool, optional): enable existing debug console output
        compresslevel (int, optional): compression level to use, between 0 and 9
        
    Returns:
        str: the output zipfile, whether we created it or determined that it already exists
    """
    
    basename = os.path.basename(input_fn)
    
    if output_fn is None:
        output_fn = input_fn + '.zip'
        
    if (not overwrite) and (os.path.isfile(output_fn)):
        print('Skipping existing file {}'.format(output_fn))
        return output_fn
    
    if verbose:
        print('Zipping {} to {} with level {}'.format(input_fn,output_fn,compresslevel))
    
    with ZipFile(output_fn,'w',zipfile.ZIP_DEFLATED) as zipf:
        zipf.write(input_fn,arcname=basename,compresslevel=compresslevel,
                   compress_type=zipfile.ZIP_DEFLATED)

    return output_fn


def add_files_to_single_tar_file(input_files, output_fn, arc_name_base,
                                 overwrite=False, verbose=False, mode='x'):
    """
    Adds all the files in [input_files] to the tar file [output_fn].  
    Archive names are relative to arc_name_base.
    
    Args:
        input_files (list): list of absolute filenames to include in the .tar file
        output_fn (str): .tar file to create
        arc_name_base (str): absolute folder from which relative paths should be determined;
            behavior is undefined if there are files in [input_files] that don't live within
            [arc_name_base]
        overwrite (bool, optional): whether to overwrite an existing .tar file
        verbose (bool, optional): enable additional debug console output
        mode (str, optional): compression type, can be 'x' (no compression), 'x:gz', or 'x:bz2'.
        
    Returns:
        str: the output tar file, whether we created it or determined that it already exists
    """
    
    if os.path.isfile(output_fn):
        if not overwrite:
            print('Tar file {} exists, skipping'.format(output_fn))
            return output_fn
        else:
            print('Tar file {} exists, deleting and re-creating'.format(output_fn))
            os.remove(output_fn)
                
    if verbose:
        print('Adding {} files to {} (mode {})'.format(
            len(input_files),output_fn,mode))
        
    with tarfile.open(output_fn,mode) as tarf:
        for input_fn_abs in tqdm(input_files,disable=(not verbose)):
            input_fn_relative = os.path.relpath(input_fn_abs,arc_name_base)
            tarf.add(input_fn_abs,arcname=input_fn_relative)

    return output_fn


def zip_files_into_single_zipfile(input_files, output_fn, arc_name_base,
                                  overwrite=False, verbose=False, compresslevel=9):
    """
    Zip all the files in [input_files] into [output_fn].  Archive names are relative to 
    arc_name_base.
    
    Args:
        input_files (list): list of absolute filenames to include in the .tar file
        output_fn (str): .tar file to create
        arc_name_base (str): absolute folder from which relative paths should be determined;
            behavior is undefined if there are files in [input_files] that don't live within
            [arc_name_base]
        overwrite (bool, optional): whether to overwrite an existing .tar file
        verbose (bool, optional): enable additional debug console output
        compresslevel (int, optional): compression level to use, between 0 and 9
        
    Returns:
        str: the output zipfile, whether we created it or determined that it already exists
    """
    
    if not overwrite:
        if os.path.isfile(output_fn):
            print('Zip file {} exists, skipping'.format(output_fn))
            return output_fn
        
    if verbose:
        print('Zipping {} files to {} (compression level {})'.format(
            len(input_files),output_fn,compresslevel))
        
    with ZipFile(output_fn,'w',zipfile.ZIP_DEFLATED) as zipf:
        for input_fn_abs in tqdm(input_files,disable=(not verbose)):
            input_fn_relative = os.path.relpath(input_fn_abs,arc_name_base)
            zipf.write(input_fn_abs,
                       arcname=input_fn_relative,
                       compresslevel=compresslevel,
                       compress_type=zipfile.ZIP_DEFLATED)

    return output_fn
    
    
def zip_folder(input_folder, output_fn=None, overwrite=False, verbose=False, compresslevel=9):
    """
    Recursively zip everything in [input_folder] into a single zipfile, storing outputs as relative 
    paths.
    
    Args: 
        input_folder (str): folder to zip
        output_fn (str, optional): output filename; if this is None, we'll write to [input_folder].zip
        overwrite (bool, optional): whether to overwrite an existing .tar file
        verbose (bool, optional): enable additional debug console output
        compresslevel (int, optional): compression level to use, between 0 and 9
        
    Returns:
        str: the output zipfile, whether we created it or determined that it already exists    
    """
    
    if output_fn is None:
        output_fn = input_folder + '.zip'
        
    if not overwrite:
        if os.path.isfile(output_fn):
            print('Zip file {} exists, skipping'.format(output_fn))
            return            
        
    if verbose:
        print('Zipping {} to {} (compression level {})'.format(
            input_folder,output_fn,compresslevel))
    
    relative_filenames = recursive_file_list(input_folder,return_relative_paths=True)
    
    with ZipFile(output_fn,'w',zipfile.ZIP_DEFLATED) as zipf:
        for input_fn_relative in tqdm(relative_filenames,disable=(not verbose)):
            input_fn_abs = os.path.join(input_folder,input_fn_relative)            
            zipf.write(input_fn_abs,
                       arcname=input_fn_relative,
                       compresslevel=compresslevel,
                       compress_type=zipfile.ZIP_DEFLATED)

    return output_fn

        
def parallel_zip_files(input_files, max_workers=16, use_threads=True, compresslevel=9, 
                       overwrite=False, verbose=False):
    """
    Zips one or more files to separate output files in parallel, leaving the 
    original files in place.  Each file is zipped to [filename].zip.
    
    Args:
        input_file (str): list of files to zip
        max_workers (int, optional): number of concurrent workers, set to <= 1 to disable parallelism
        use_threads (bool, optional): whether to use threads (True) or processes (False); ignored if
            max_workers <= 1
        compresslevel (int, optional): zip compression level between 0 and 9
        overwrite (bool, optional): whether to overwrite an existing .tar file
        verbose (bool, optional): enable additional debug console output
    """

    n_workers = min(max_workers,len(input_files))

    if use_threads:
        pool = ThreadPool(n_workers)
    else:
        pool = Pool(n_workers)

    with tqdm(total=len(input_files)) as pbar:
        for i,_ in enumerate(pool.imap_unordered(partial(zip_file,
          output_fn=None,overwrite=overwrite,verbose=verbose,compresslevel=compresslevel),
          input_files)):
            pbar.update()


def parallel_zip_folders(input_folders, max_workers=16, use_threads=True,
                         compresslevel=9, overwrite=False, verbose=False):
    """
    Zips one or more folders to separate output files in parallel, leaving the 
    original folders in place.  Each folder is zipped to [folder_name].zip.
    
    Args:
        input_folder (list): list of folders to zip
        max_workers (int, optional): number of concurrent workers, set to <= 1 to disable parallelism
        use_threads (bool, optional): whether to use threads (True) or processes (False); ignored if
            max_workers <= 1
        compresslevel (int, optional): zip compression level between 0 and 9
        overwrite (bool, optional): whether to overwrite an existing .tar file
        verbose (bool, optional): enable additional debug console output
    """

    n_workers = min(max_workers,len(input_folders))

    if use_threads:
        pool = ThreadPool(n_workers)
    else:
        pool = Pool(n_workers)
    
    with tqdm(total=len(input_folders)) as pbar:
        for i,_ in enumerate(pool.imap_unordered(
                partial(zip_folder,overwrite=overwrite,
                        compresslevel=compresslevel,verbose=verbose),
                input_folders)):
            pbar.update()


def zip_each_file_in_folder(folder_name,recursive=False,max_workers=16,use_threads=True,
                            compresslevel=9,overwrite=False,required_token=None,verbose=False,
                            exclude_zip=True):
    """
    Zips each file in [folder_name] to its own zipfile (filename.zip), optionally recursing.  To 
    zip a whole folder into a single zipfile, use zip_folder().
    
    Args:
        folder_name (str): the folder within which we should zip files
        recursive (bool, optional): whether to recurse within [folder_name]
        max_workers (int, optional): number of concurrent workers, set to <= 1 to disable parallelism
        use_threads (bool, optional): whether to use threads (True) or processes (False); ignored if
            max_workers <= 1
        compresslevel (int, optional): zip compression level between 0 and 9
        overwrite (bool, optional): whether to overwrite an existing .tar file
        required_token (str, optional): only zip files whose names contain this string
        verbose (bool, optional): enable additional debug console output
        exclude_zip (bool, optional): skip files ending in .zip        
    """
    
    assert os.path.isdir(folder_name), '{} is not a folder'.format(folder_name)
    
    input_files = recursive_file_list(folder_name,recursive=recursive,return_relative_paths=False)
    
    if required_token is not None:
        input_files = [fn for fn in input_files if required_token in fn]
    
    if exclude_zip:
        input_files = [fn for fn in input_files if (not fn.endswith('.zip'))]
                                                    
    parallel_zip_files(input_files=input_files,max_workers=max_workers,
                       use_threads=use_threads,compresslevel=compresslevel,
                       overwrite=overwrite,verbose=verbose)


def unzip_file(input_file, output_folder=None):
    """
    Unzips a zipfile to the specified output folder, defaulting to the same location as
    the input file.
    
    Args:
        input_file (str): zipfile to unzip
        output_folder (str, optional): folder to which we should unzip [input_file], defaults
            to unzipping to the folder where [input_file] lives
    """
    
    if output_folder is None:
        output_folder = os.path.dirname(input_file)
        
    with zipfile.ZipFile(input_file, 'r') as zf:
        zf.extractall(output_folder)



"""

 bellevue_to_json.py

 "Bellevue Camera Traps" is the rather unremarkable camera trap data set
 used by one of the repo's maintainers for testing.  It's organized as:

 approximate_date/[loose_camera_specifier/]/species    

 E.g.:
    
 "2018.03.30\coyote\DSCF0091.JPG"
 "2018.07.18\oldcam\empty\DSCF0001.JPG"

"""

#%% Constants and imports

import json
import os
import uuid
import datetime

from PIL import Image
from PIL.ExifTags import TAGS
from tqdm import tqdm

from megadetector.utils.path_utils import find_images

# Filenames will be stored in the output .json relative to this base dir
base_dir = r'C:\temp\camera_trap_images_no_people'
output_base = r'c:\temp\previews'
output_filename = os.path.join(base_dir,'bellevue_camera_traps.{}.json'.format(str(datetime.date.today())))

class_mappings = {'transitional':'unlabeled','moving':'unlabeled','setup':'unlabeled','blurry':'unlabeled','transitional':'unlabeled','junk':'unlabeled','unknown':'unlabeled','blurry':'unlabeled'}
class_mappings['dan'] = 'human'
class_mappings['dan_and_dog'] = 'human,dog'
class_mappings['dan and dog'] = 'human,dog'
class_mappings['unknown'] = 'unknown animal'
class_mappings['racoon'] = 'raccoon'


info = {}
info['year'] = 2020
info['version'] = '2.0'
info['description'] = 'Bellevue Camera Traps'
info['contributor'] = 'Dan Morris'
info['date_created'] = str(datetime.date.today())

max_files = -1

max_seconds_within_sequence = 10.0

assert os.path.isdir(base_dir)

#%% Exif functions

def get_exif_tags(fn=None,im=None):
    
    assert (fn is not None) or (im is not None)
    ret = {}
    if im is None:
        im = Image.open(fn)
    info = im._getexif()
    for tag, value in info.items():
        decoded = TAGS.get(tag, tag)
        ret[decoded] = value
        
    return ret


#%% Enumerate files, create image/annotation/category info

annotations = []

category_name_to_category = {}

# Force the empty category to be ID 0
empty_category = {}
empty_category['id'] = 0
empty_category['name'] = 'empty'
category_name_to_category['empty'] = empty_category
next_category_id = 1

# Keep track of unique camera folders
camera_folders = set()

# Each element will be a dictionary with fields:
#
# relative_path, width, height, datetime    
images = []
non_image_files = []

print('Enumerating files from {}'.format(base_dir))

image_files = find_images(base_dir,recursive=True)
print('Enumerated {} images'.format(len(image_files)))

# fname = image_files[0]
for fname in tqdm(image_files):
  
    if max_files >= 0 and len(images) > max_files:            
        print('Warning: early break at {} files'.format(max_files))
        break
    
    full_path = fname
    relative_path = os.path.relpath(full_path,base_dir)
     
    try:
        im = Image.open(full_path)
        h = im.height
        w = im.width
        tags = get_exif_tags(None,im)
        s = tags['DateTimeOriginal']
        dt = datetime.datetime.strptime(s,'%Y:%m:%d %H:%M:%S')
    except:
        # Corrupt or not an image
        print('Warning: could not read {}'.format(fname))
        non_image_files.append(full_path)
        continue
    
    # Store file info
    image_info = {}
    image_info['file_name'] = relative_path
    image_info['width'] = w
    image_info['height'] = h
    image_info['datetime'] = dt
    image_info['location'] = 'unknown'
    image_info['id'] = str(uuid.uuid4())

    images.append(image_info)
    
    # E.g. 2018.03.30/coyote/DSCF0091.JPG
    relative_path = image_info['file_name'].replace('\\','/')
    tokens = relative_path.split('/')
    camera_path_tokens = tokens[0:-2]
    camera_path = '/'.join(camera_path_tokens)
    camera_folders.add(camera_path)
    image_info['camera_path'] = camera_path
    
    category_name = tokens[-2].lower()
    if category_name in class_mappings:
        category_name = class_mappings[category_name]
        
    if category_name not in category_name_to_category:
        category = {}
        category['id'] = next_category_id
        category['name'] = category_name
        next_category_id = next_category_id + 1
        category_name_to_category[category_name] = category
    else:
        category = category_name_to_category[category_name]
    
    annotation = {}
    annotation['sequence_level_annotation'] = False
    annotation['id'] = str(uuid.uuid4())
    annotation['category_id'] = category['id']
    annotation['image_id'] = image_info['id']
    annotations.append(annotation)
        
# ...for each image file

assert len(annotations) == len(images)

categories = list(category_name_to_category.values())


#%% Synthesize sequence information        

print('Found {} camera folders'.format(len(camera_folders)))

camera_folders = list(camera_folders)

all_sequences = set()

# Sort images by time within each folder
# camera_path = camera_folders[0]
for i_camera,camera_path in enumerate(camera_folders):
    
    images_this_camera = [im for im in images if im['camera_path'] == camera_path]
    sorted_images_this_camera = sorted(images_this_camera, key = lambda im: im['datetime'])
    
    current_sequence_id = None
    next_sequence_index = 0
    previous_datetime = None
        
    # previous_datetime = sorted_images_this_camera[0]['datetime']
    # im = sorted_images_this_camera[1]
    for im in sorted_images_this_camera:
        
        if previous_datetime is None:
            delta = None
        else:
            delta = (im['datetime'] - previous_datetime).total_seconds()
        
        # Start a new sequence if necessary
        if delta is None or delta > max_seconds_within_sequence:
            next_sequence_index = 0
            current_sequence_id = str(uuid.uuid4())
            all_sequences.add(current_sequence_id)
            
        im['seq_id'] = current_sequence_id
        im['seq_num_frames'] = None
        im['frame_num'] = next_sequence_index
        next_sequence_index = next_sequence_index + 1
        previous_datetime = im['datetime']
    
    # ...for each image in this camera

# ...for each camera

print('Created {} sequences from {} images'.format(len(all_sequences),len(images)))

# Fill in seq_num_frames
num_frames_per_sequence = {}
for seq_id in all_sequences:
    images_this_sequence = [im for im in images if im['seq_id'] == seq_id]
    num_frames_per_sequence[seq_id] = len(images_this_sequence)
    for im in images_this_sequence:
        im['seq_num_frames'] = len(images_this_sequence)


#%% A little cleanup 

for im in tqdm(images):
    if 'camera_path' in im:
        del im['camera_path']
    if not isinstance(im['datetime'],str):
        im['datetime'] = str(im['datetime'])
    
    
#%% Write output .json

data = {}
data['info'] = info
data['images'] = images
data['annotations'] = annotations
data['categories'] = categories

json.dump(data, open(output_filename,'w'), indent=1)

print('Finished writing json to {}'.format(output_filename))


#%% Validate data

from megadetector.data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = base_dir
options.bCheckImageSizes = False
options.bCheckImageExistence = True
options.bFindUnusedImages = False
    
sorted_categories  = integrity_check_json_db.integrity_check_json_db(output_filename,options)


#%% Label previews

from megadetector.visualization import visualize_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = None
viz_options.parallelize_rendering_n_cores = 8
viz_options.parallelize_rendering = True
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = True
viz_options.sort_by_filename = False
html_output_file,image_db = visualize_db.visualize_db(output_filename,
                                                        os.path.join(output_base,'preview'),
                                                        base_dir,viz_options)
os.startfile(html_output_file)




"""

 noaa_seals_2019.py

 Prepare the NOAA Arctic Seals 2019 metadata for LILA.

"""

#%% Imports and constants

import os
import pandas as pd
from tqdm import tqdm

from megadetector.utils import url_utils
from megadetector.visualization import visualization_utils

# A list of files in the lilablobssc container for this data set
container_file_list = r'C:\temp\seals\seal_files.txt'

# The raw detection files provided by NOAA
detections_fn = r'C:\temp\seals\surv_test_kamera_detections_20210212.csv'

# A version of the above with filename columns added
detections_fn_full_paths = detections_fn.replace('.csv','_full_paths.csv')

base_url = 'https://lilablobssc.blob.core.windows.net/noaa-kotz'


#%% Read input .csv

df = pd.read_csv(detections_fn)
df['rgb_image_path'] = ''
df['ir_image_path'] = ''
print('Read {} rows from {}'.format(len(df),detections_fn))

camera_view_to_path = {}
camera_view_to_path['C'] = 'CENT'
camera_view_to_path['L'] = 'LEFT'

valid_flights = set(['fl04','fl05','fl06','fl07'])


#%% Read list of files

with open(container_file_list,'r') as f:
    all_files = f.readlines()
all_files = [s.strip() for s in all_files]
all_files = set(all_files)


#%% Convert paths to full paths

missing_ir_files = []

# i_row = 0; row = df.iloc[i_row]
for i_row,row in tqdm(df.iterrows(),total=len(df)):
    
    assert row['flight'] in valid_flights
    assert row['camera_view'] in camera_view_to_path
    
    assert isinstance(row['rgb_image_name'],str)
    rgb_image_path = 'Images/{}/{}/{}'.format(row['flight'],camera_view_to_path[row['camera_view']],
                                     row['rgb_image_name'])
    assert rgb_image_path in all_files
    df.loc[i_row,'rgb_image_path'] = rgb_image_path
    
    if not isinstance(row['ir_image_name'],str):
        continue
    
    ir_image_path = 'Images/{}/{}/{}'.format(row['flight'],camera_view_to_path[row['camera_view']],
                                     row['ir_image_name'])
    # assert ir_image_path in all_files
    if ir_image_path not in all_files:
        missing_ir_files.append(ir_image_path)
    df.loc[i_row,'ir_image_path'] = ir_image_path
            
# ...for each row    

missing_ir_files = list(set(missing_ir_files))
missing_ir_files.sort()
print('{} missing IR files (of {})'.format(len(missing_ir_files),len(df)))

for s in missing_ir_files:
    print(s)


#%% Write results

df.to_csv(detections_fn_full_paths,index=False)


#%% Load output file, just to be sure

df = pd.read_csv(detections_fn_full_paths)


#%% Render annotations on an image

import random; i_image = random.randint(0,len(df))
# i_image = 2004 
row = df.iloc[i_image]
rgb_image_path = row['rgb_image_path']
rgb_image_url = base_url + '/' + rgb_image_path
ir_image_path = row['ir_image_path']
ir_image_url = base_url + '/' + ir_image_path


#%% Download the image

rgb_image_fn = url_utils.download_url(rgb_image_url,progress_updater=True)
ir_image_fn = url_utils.download_url(ir_image_url,progress_updater=True)


#%% Find all the rows (detections) associated with this image

# as l,r,t,b
rgb_boxes = []
ir_boxes = []

for i_row,row in df.iterrows():
    
    if row['rgb_image_path'] == rgb_image_path:
        box_l = row['rgb_left']
        box_r = row['rgb_right']
        box_t = row['rgb_top']
        box_b = row['rgb_bottom']
        rgb_boxes.append([box_l,box_r,box_t,box_b])

    if row['ir_image_path'] == ir_image_path:
        box_l = row['ir_left']
        box_r = row['ir_right']
        box_t = row['ir_top']
        box_b = row['ir_bottom']
        ir_boxes.append([box_l,box_r,box_t,box_b])
        
print('Found {} RGB, {} IR annotations for this image'.format(len(rgb_boxes),
                                                              len(ir_boxes))) 


#%% Render the detections on the image(s)

img_rgb = visualization_utils.load_image(rgb_image_fn)
img_ir = visualization_utils.load_image(ir_image_fn)

for b in rgb_boxes:
    
    # In pixel coordinates
    box_left = b[0]; box_right = b[1]; box_top = b[2]; box_bottom = b[3]    
    assert box_top > box_bottom; assert box_right > box_left    
    ymin = box_bottom; ymax = box_top; xmin = box_left; xmax = box_right

    visualization_utils.draw_bounding_box_on_image(img_rgb,ymin,xmin,ymax,xmax,
                                                   use_normalized_coordinates=False,
                                                   thickness=3)

for b in ir_boxes:
    
    # In pixel coordinates
    box_left = b[0]; box_right = b[1]; box_top = b[2]; box_bottom = b[3]    
    assert box_top > box_bottom; assert box_right > box_left    
    ymin = box_bottom; ymax = box_top; xmin = box_left; xmax = box_right

    visualization_utils.draw_bounding_box_on_image(img_ir,ymin,xmin,ymax,xmax,
                                                   use_normalized_coordinates=False,
                                                   thickness=3)

# visualization_utils.show_images_in_a_row([img_rgb,img_ir])


#%% Save images

img_rgb.save(r'c:\temp\seals_rgb.png')
img_ir.save(r'c:\temp\seals_ir.png')


#%% Clean up

import shutil
tmp_dir = os.path.dirname(rgb_image_fn)
shutil.rmtree(tmp_dir)



"""

ocr_tools.py

Use OCR (via the Tesseract package) to pull metadata (particularly times and
dates from camera trap images).

The general approach is:

* Crop a fixed percentage from the top and bottom of an image, slightly larger
  than the largest examples we've seen of how much space is used for metadata.

* Define the background color as the median pixel value, and find rows that are
  mostly that color to refine the crop.

* Crop to the refined crop, then run pytesseract to extract text.

* Use regular expressions to find time and date.

Prior to using this module:

* Install Tesseract from https://tesseract-ocr.github.io/tessdoc/Installation.html

* pip install pytesseract
   
Known limitations:

* Semi-transparent overlays (which I've only seen on consumer cameras) usually fail.
   
"""

#%% Notes to self

"""

* To use the legacy engine (--oem 0), I had to download an updated eng.traineddata file from:
    
  https://github.com/tesseract-ocr/tessdata
  
"""

#%% Constants and imports

import os
import json
import numpy as np
import datetime
import re

from functools import partial
from dateutil.parser import parse as dateparse

import cv2
from PIL import Image, ImageFilter
from tqdm import tqdm

from megadetector.utils.path_utils import find_images
from megadetector.utils.path_utils import open_file
from megadetector.utils import write_html_image_list        
from megadetector.utils.ct_utils import is_iterable
from megadetector.visualization import visualization_utils as vis_utils

# pip install pytesseract
#
# Also install tesseract from: https://github.com/UB-Mannheim/tesseract/wiki, and add
# the installation dir to your path (on Windows, typically C:\Program Files (x86)\Tesseract-OCR)
import pytesseract


#%% Extraction options

class DatetimeExtractionOptions:
    """
    Options used to parameterize datetime extraction in most functions in this module.
    """
    
    def __init__(self):
        
        #: Using a semi-arbitrary metric of how much it feels like we found the 
        #: text-containing region, discard regions that appear to be extraction failures
        self.p_crop_success_threshold = 0.5
        
        #: Pad each crop with a few pixels to make tesseract happy
        self.crop_padding = 10        
        
        #: Discard short text, typically text from the top of the image
        self.min_text_length = 4
        
        #: When we're looking for pixels that match the background color, allow some 
        #: tolerance around the dominant color
        self.background_tolerance = 2
            
        #: We need to see a consistent color in at least this fraction of pixels in our rough 
        #: crop to believe that we actually found a candidate metadata region.
        self.min_background_fraction = 0.3
        
        #: What fraction of the [top,bottom] of the image should we use for our rough crop?
        self.image_crop_fraction = [0.045 , 0.045]
        # self.image_crop_fraction = [0.08 , 0.08]
        
        #: Within that rough crop, how much should we use for determining the background color?
        self.background_crop_fraction_of_rough_crop = 0.5
        
        #: A row is considered a probable metadata row if it contains at least this fraction
        #: of the background color.  This is used only to find the top and bottom of the crop area, 
        #: so it's not that *every* row needs to hit this criteria, only the rows that are generally
        #: above and below the text.
        self.min_background_fraction_for_background_row = 0.5
        
        #: psm 6: "assume a single uniform block of text"
        #: psm 13: raw line
        #: oem: 0 == legacy, 1 == lstm
        #: tesseract_config_string = '--oem 0 --psm 6'
        #:
        #: Try these configuration strings in order until we find a valid datetime
        self.tesseract_config_strings = ['--oem 1 --psm 13','--oem 0 --psm 13',
                                         '--oem 1 --psm 6','--oem 0 --psm 6']
        
        #: If this is False, and one set of options appears to succeed for an image, we'll
        #: stop there.  If this is True, we always run all option sets on every image.
        self.force_all_ocr_options = False
        
        #: Whether to apply PIL's ImageFilter.SHARPEN prior to OCR
        self.apply_sharpening_filter = True
        
        #: Tesseract should be on your system path, but you can also specify the
        #: path explicitly, e.g. you can do either of these:
        #:
        #: * os.environ['PATH'] += r';C:\Program Files\Tesseract-OCR'
        #: * self.tesseract_cmd = 'r"C:\Program Files\Tesseract-OCR\tesseract.exe"'
        self.tesseract_cmd = 'tesseract.exe'


#%% Support functions

def make_rough_crops(image,options=None):
    """
    Crops the top and bottom regions out of an image.
    
    Args:
        image (Image or str): a PIL Image or file name
        options (DatetimeExtractionOptions, optional): OCR parameters
        
    Returns:
        dict: a dict with fields 'top' and 'bottom', each pointing to a new PIL Image        
    """
    
    if options is None:
        options = DatetimeExtractionOptions()
        
    if isinstance(image,str):
        image = vis_utils.open_image(image)
        
    w = image.width
    h = image.height
    
    crop_height_top = round(options.image_crop_fraction[0] * h)
    crop_height_bottom = round(options.image_crop_fraction[1] * h)
    
    # l,t,r,b
    #
    # 0,0 is upper-left
    top_crop = image.crop([0,0,w,crop_height_top])
    bottom_crop = image.crop([0,h-crop_height_bottom,w,h])
    return {'top':top_crop,'bottom':bottom_crop}
    
# ...def make_rough_crops(...)


def crop_to_solid_region(rough_crop,crop_location,options=None):
    """       
    Given a rough crop from the top or bottom of an image, finds the background color
    and crops to the metadata region.
    
    Within a region of an image (typically a crop from the top-ish or bottom-ish part of 
    an image), tightly crop to the solid portion (typically a region with a black background).

    The success metric is just a binary indicator right now: 1.0 if we found a region we believe
    contains a solid background, 0.0 otherwise.
    
    Args:
        rough_crop (Image): the PIL Image to crop
        crop_location (str): 'top' or 'bottom'
        options (DatetimeExtractionOptions, optional): OCR parameters
        
    Returns:
        tuple: a tuple containing (a cropped_image (Image), p_success (float), padded_image (Image))
    """
    
    if options is None:
        options = DatetimeExtractionOptions()    

    crop_to_solid_region_result = {}
    crop_to_solid_region_result['crop_pil'] = None
    crop_to_solid_region_result['padded_crop_pil'] = None
    crop_to_solid_region_result['p_success'] = 0.0
    
    # pil --> cv2        
    rough_crop_np = np.array(rough_crop) 
    rough_crop_np = rough_crop_np[:, :, ::-1].copy()         
    
    # Search *part* of the crop for the background value (the part closest to the top or bottom
    # of the image)
    rows_to_use_for_background_search = int(rough_crop_np.shape[0] * \
                                            options.background_crop_fraction_of_rough_crop)
    
    if crop_location == 'top':
        background_search_image = rough_crop_np[0:rows_to_use_for_background_search,:,:]
    elif crop_location == 'bottom':
        background_search_image = rough_crop_np[-rows_to_use_for_background_search:,:,:]
    else:
        raise ValueError('Unrecognized crop location: {}'.format(crop_location))
                
    background_search_image = cv2.cvtColor(background_search_image, cv2.COLOR_BGR2GRAY)
    background_search_image = background_search_image.astype('uint8')     
    background_search_image = cv2.medianBlur(background_search_image,3) 
    pixel_values = background_search_image.flatten()
    counts = np.bincount(pixel_values)
    background_value = int(np.argmax(counts))
    
    # Did we find a sensible mode that looks like a background value?
    background_value_count = int(np.max(counts))
    p_background_value = background_value_count / np.sum(counts)
    
    if (p_background_value < options.min_background_fraction):
        return crop_to_solid_region_result
    else:
        p_success = 1.0
        
    analysis_image = cv2.cvtColor(rough_crop_np, cv2.COLOR_BGR2GRAY)
    analysis_image = analysis_image.astype('uint8')     
    analysis_image = cv2.medianBlur(analysis_image,3) 
    
    # This will now be a binary image indicating which pixels are background
    analysis_image = cv2.inRange(analysis_image,
                                background_value-options.background_tolerance,
                                background_value+options.background_tolerance)
    
    # Use row heuristics to refine the crop        
    h = analysis_image.shape[0]
    w = analysis_image.shape[1]
    
    min_x = 0
    min_y = -1
    max_x = w
    max_y = -1
    
    # Find the first and last row that are mostly the background color
    for y in range(h):
        row_count = 0
        for x in range(w):
            if analysis_image[y][x] > 0:
                row_count += 1
        row_fraction = row_count / w
        if row_fraction > options.min_background_fraction_for_background_row:
            if min_y == -1:
                min_y = y
            max_y = y
    
    assert (min_y == -1 and max_y == -1) or (min_y != -1 and max_y != -1)
    
    if min_y == -1:
        return crop_to_solid_region_result
    
    if max_y == min_y:
        return crop_to_solid_region_result
    
    x = min_x
    y = min_y
    w = max_x-min_x
    h = max_y-min_y
    
    x = min_x
    y = min_y
    w = max_x-min_x
    h = max_y-min_y

    # Crop the image
    crop_np = rough_crop_np[y:y+h,x:x+w]
      
    # Tesseract doesn't like characters really close to the edge, so pad a little.
    crop_padding = options.crop_padding
    padded_crop_np = cv2.copyMakeBorder(crop_np,crop_padding,crop_padding,crop_padding,crop_padding,
                                     cv2.BORDER_CONSTANT,
                                     value=[background_value,background_value,background_value])

    crop_pil = Image.fromarray(crop_np)
    padded_crop_pil = Image.fromarray(padded_crop_np)
    
    crop_to_solid_region_result['crop_pil'] = crop_pil
    crop_to_solid_region_result['padded_crop_pil'] = padded_crop_pil
    crop_to_solid_region_result['p_success'] = p_success
    
    return crop_to_solid_region_result
    
# ...crop_to_solid_region(...)    


def find_text_in_crops(rough_crops,options=None,tesseract_config_string=None):
    """
    Finds all text in each Image in the dict [rough_crops]; those images should be pretty small 
    regions by the time they get to this function, roughly the top or bottom 20% of an image.
    
    Args:
        rough_crops (list): list of Image objects that have been cropped close to text
        options (DatetimeExtractionOptions, optional): OCR parameters
        tesseract_config_string (str, optional): optional CLI argument to pass to tesseract.exe
        
    Returns:
        dict: a dict with keys "top" and "bottom", where each value is a dict with keys
        'text' (text found, if any) and 'crop_to_solid_region_results' (metadata about the OCR pass)
    """
    
    if options is None:
        options = DatetimeExtractionOptions()
    
    if tesseract_config_string is None:
        tesseract_config_string = options.tesseract_config_strings[0]
        
    find_text_in_crops_results = {}
    
    # crop_location = 'top'
    # crop_location = 'bottom'
    for crop_location in ('top','bottom'):

        find_text_in_crops_results[crop_location] = {}
        find_text_in_crops_results[crop_location]['text'] = ''
        find_text_in_crops_results[crop_location]['crop_to_solid_region_results'] = None
        
        rough_crop = rough_crops[crop_location]
        
        # Crop to the portion of the rough crop with a solid background color
        crop_to_solid_region_results = crop_to_solid_region(rough_crop,crop_location,options)
        
        find_text_in_crops_results[crop_location]['crop_to_solid_region_results'] = \
            crop_to_solid_region_results
        
        # Try cropping to a solid region; if that doesn't work, try running OCR on the whole
        # rough crop.
        if crop_to_solid_region_results['p_success'] >= options.p_crop_success_threshold:
            padded_crop_pil = crop_to_solid_region_results['padded_crop_pil']
        else:            
            # continue
            padded_crop_pil = rough_crop        
            
        if options.apply_sharpening_filter:
            padded_crop_pil = padded_crop_pil.filter(ImageFilter.SHARPEN)
        
        # Find text in the padded crop
        pytesseract.pytesseract.tesseract_cmd = options.tesseract_cmd
        text = pytesseract.image_to_string(padded_crop_pil, lang='eng', 
                                           config=tesseract_config_string)
        
        text = text.replace('\n', ' ').replace('\r', '').strip()

        find_text_in_crops_results[crop_location]['text'] = text        
                
    # ...for each cropped region
    
    return find_text_in_crops_results
    
# ...def find_text_in_crops(...)
    

def _datetime_string_to_datetime(matched_string):
    """
    Takes an OCR-matched datetime string, does a little cleanup, and parses a date
    from it.
    
    By the time a string gets to this function, it should be a proper date string, with
    no extraneous characters other than spaces around colons or hyphens.
    """
    
    matched_string = matched_string.replace(' -','-')
    matched_string = matched_string.replace('- ','-')
    matched_string = matched_string.replace(' :',':')
    matched_string = matched_string.replace(': ',':')
    try:
        extracted_datetime = dateparse(matched_string)
    except Exception:
        extracted_datetime = None
    return extracted_datetime


def _get_datetime_from_strings(strings,options=None):
    """
    Given a string or list of strings, search for exactly one datetime in those strings.
    using a series of regular expressions.
    
    Strings are currently just concatenated before searching for a datetime.
    """
    
    if options is None:
        options = DatetimeExtractionOptions()    
    
    if isinstance(strings,str):
        s = strings
    else:
        s = ' '.join(strings).lower()
    s = s.replace('—','-')    
    s = ''.join(e for e in s if e.isalnum() or e in ':-/' or e.isspace())
        
    ### AM/PM
    
    # 2013-10-02 11:40:50 AM
    m = re.search('(\d\d\d\d)\s?-\s?(\d\d)\s?-\s?(\d\d)\s+(\d+)\s?:?\s?(\d\d)\s?:\s?(\d\d)\s*([a|p]m)',s)
    if m is not None:        
        return _datetime_string_to_datetime(m.group(0))        
    
    # 04/01/2017 08:54:00AM
    m = re.search('(\d\d)\s?/\s?(\d\d)\s?/\s?(\d\d\d\d)\s+(\d+)\s?:\s?(\d\d)\s?:\s?(\d\d)\s*([a|p]m)',s)
    if m is not None:        
        return _datetime_string_to_datetime(m.group(0))    
    
    # 2017/04/01 08:54:00AM
    m = re.search('(\d\d\d\d)\s?/\s?(\d\d)\s?/\s?(\d\d)\s+(\d+)\s?:\s?(\d\d)\s?:\s?(\d\d)\s*([a|p]m)',s)
    if m is not None:        
        return _datetime_string_to_datetime(m.group(0))        
    
    # 04/01/2017 08:54AM
    m = re.search('(\d\d)\s?/\s?(\d\d)\s?/\s?(\d\d\d\d)\s+(\d+)\s?:\s?(\d\d)\s*([a|p]m)',s)
    if m is not None:        
        return _datetime_string_to_datetime(m.group(0))    
    
    # 2017/04/01 08:54AM
    m = re.search('(\d\d\d\d)\s?/\s?(\d\d)\s?/\s?(\d\d)\s+(\d+)\s?:\s?(\d\d)\s*([a|p]m)',s)
    if m is not None:        
        return _datetime_string_to_datetime(m.group(0))        
    
    ### No AM/PM
    
    # 2013-07-27 04:56:35
    m = re.search('(\d\d\d\d)\s?-\s?(\d\d)\s?-\s?(\d\d)\s*(\d\d)\s?:\s?(\d\d)\s?:\s?(\d\d)',s)
    if m is not None:        
        return _datetime_string_to_datetime(m.group(0))        
    
    # 07-27-2013 04:56:35
    m = re.search('(\d\d)\s?-\s?(\d\d)\s?-\s?(\d\d\d\d)\s*(\d\d)\s?:\s?(\d\d)\s?:\s?(\d\d)',s)
    if m is not None:        
        return _datetime_string_to_datetime(m.group(0))        
    
    # 2013/07/27 04:56:35
    m = re.search('(\d\d\d\d)\s?/\s?(\d\d)\s?/\s?(\d\d)\s*(\d\d)\s?:\s?(\d\d)\s?:\s?(\d\d)',s)
    if m is not None:        
        return _datetime_string_to_datetime(m.group(0))        
    
    # 07/27/2013 04:56:35
    m = re.search('(\d\d)\s?/\s?(\d\d)\s?/\s?(\d\d\d\d)\s*(\d\d)\s?:\s?(\d\d)\s?:\s?(\d\d)',s)
    if m is not None:        
        return _datetime_string_to_datetime(m.group(0))        
    
    return None
    
# ...def _get_datetime_from_strings(...)


def get_datetime_from_image(image,include_crops=True,options=None):
    """
    Tries to find the datetime string (if present) in an image.
    
    Args:
        image (Image or str): the PIL Image object or image filename in which we should look for
            datetime information.
        include_crops (bool, optional): whether to include cropped images in the return dict (set
            this to False if you're worried about size and you're processing a zillion images)
        options (DatetimeExtractionOptions or list, optional): OCR parameters, either one 
            DatetimeExtractionOptions object or a list of options to try
    
    Returns:
        dict: a dict with fields:
            
            - datetime: Python datetime object, or None
            - text_results: length-2 list of strings
            - all_extracted_datetimes: if we ran multiple option sets, this will contain the 
              datetimes extracted for each option set
            - ocr_results: detailed results from the OCR process, including crops as PIL images;
              only included if include_crops is True
    """
    
    if options is None:
        options = DatetimeExtractionOptions()
    
    if isinstance(image,str):
        image = vis_utils.open_image(image)

    # Crop the top and bottom from the image
    rough_crops = make_rough_crops(image,options)
    assert len(rough_crops) == 2
    
    all_extracted_datetimes = {}
    all_text_results = []
    all_ocr_results = []
    
    extracted_datetime = None
    
    # Find text, possibly trying all config strings
    #
    # tesseract_config_string = options.tesseract_config_strings[0]
    for tesseract_config_string in options.tesseract_config_strings:
        
        ocr_results = find_text_in_crops(rough_crops,options,tesseract_config_string)
        all_ocr_results.append(ocr_results)
        
        text_results = [v['text'] for v in ocr_results.values()]
        assert len(text_results) == 2
        all_text_results.append(text_results)
            
        # Find datetime
        extracted_datetime_this_option_set = _get_datetime_from_strings(text_results,options)
        assert isinstance(extracted_datetime_this_option_set,datetime.datetime) or \
            (extracted_datetime_this_option_set is None)
        
        all_extracted_datetimes[tesseract_config_string] = \
            extracted_datetime_this_option_set
            
        if extracted_datetime_this_option_set is not None:
            if extracted_datetime is None:
                extracted_datetime = extracted_datetime_this_option_set
            if not options.force_all_ocr_options:
                break        
            
    # ...for each set of OCR options
    
    if extracted_datetime is not None:        
        assert extracted_datetime.year <= 2023 and extracted_datetime.year >= 1990

    to_return = {}
    to_return['datetime'] = extracted_datetime
    
    to_return['text_results'] = all_text_results
    to_return['all_extracted_datetimes'] = all_extracted_datetimes
    
    if include_crops:
        to_return['ocr_results'] = all_ocr_results
    else:
        to_return['ocr_results'] = None
        
    return to_return

# ...def get_datetime_from_image(...)


def try_get_datetime_from_image(filename,include_crops=False,options=None):
    """
    Try/catch wrapper for get_datetime_from_image, optionally trying multiple option sets
    until we find a datetime.
    
    Args:
        image (Image or str): the PIL Image object or image filename in which we should look for
            datetime information.
        include_crops (bool, optional): whether to include cropped images in the return dict (set
            this to False if you're worried about size and you're processing a zillion images)
        options (DatetimeExtractionOptions or list, optional): OCR parameters, either one 
            DatetimeExtractionOptions object or a list of options to try
    
    Returns:
        dict: A dict with fields:
            - datetime: Python datetime object, or None
            - text_results: length-2 list of strings
            - all_extracted_datetimes: if we ran multiple option sets, this will contain the 
              datetimes extracted for each option set
            - ocr_results: detailed results from the OCR process, including crops as PIL images;
              only included if include_crops is True
    """
    
    if options is None:
        options = DatetimeExtractionOptions()

    if not is_iterable(options):
        options = [options]
    
    result = {}
    result['error'] = None
    
    for i_option_set,current_options in enumerate(options):
        try:
            result = get_datetime_from_image(filename,include_crops=include_crops,options=current_options)
            result['options_index'] = i_option_set
            if 'datetime' in result and result['datetime'] is not None:
                break
        except Exception as e:
            result['error'] = str(e)
    
    return result


def get_datetimes_for_folder(folder_name,output_file=None,n_to_sample=-1,options=None,
                             n_workers=16,use_threads=False):
    """
    The main entry point for this module.  Tries to retrieve metadata from pixels for every 
    image in [folder_name], optionally the results to the .json file [output_file].
    
    Args:
        folder_name (str): the folder of images to process recursively
        output_file (str, optional): the .json file to which we should write results; if None,
            just returns the results
        n_to_sample (int, optional): for debugging only, used to limit the number of images
            we process
        options (DatetimeExtractionOptions or list, optional): OCR parameters, either one 
            DatetimeExtractionOptions object or a list of options to try for each image
        n_workers (int, optional): the number of parallel workers to use; set to <= 1 to disable
            parallelization
        use_threads (bool, optional): whether to use threads (True) or processes (False) for
            parallelization; not relevant if n_workers <= 1
            
    Returns:
        dict: a dict mapping filenames to datetime extraction results, see try_get_datetime_from_images
        for the format of each value in the dict.
    """
    
    if options is None:
        options = DatetimeExtractionOptions()
    
    image_file_names = \
        find_images(folder_name,convert_slashes=True,
                    return_relative_paths=False,recursive=True)
    
    if n_to_sample > 0:
        import random
        random.seed(0)
        image_file_names = random.sample(image_file_names,n_to_sample)
            
    if n_workers <= 1:
        
        all_results = []
        for fn_abs in tqdm(image_file_names):
            all_results.append(try_get_datetime_from_image(fn_abs,options=options))
            
    else:    
        
        # Don't spawn more than one worker per image
        if n_workers > len(image_file_names):
            n_workers = len(image_file_names)
            
        if use_threads:
            from multiprocessing.pool import ThreadPool
            pool = ThreadPool(n_workers)
            worker_string = 'threads'        
        else:
            from multiprocessing.pool import Pool
            pool = Pool(n_workers)
            worker_string = 'processes'
            
        print('Starting a pool of {} {}'.format(n_workers,worker_string))
        
        all_results = list(tqdm(pool.imap(
            partial(try_get_datetime_from_image,options=options),image_file_names),
            total=len(image_file_names)))
    
    filename_to_results = {}
    
    # fn_relative = image_file_names[0]
    for i_file,fn_abs in enumerate(image_file_names):
        filename_to_results[fn_abs] = all_results[i_file]
    
    if output_file is not None:
        with open(output_file,'w') as f:
            json.dump(filename_to_results,f,indent=1,default=str)

    return filename_to_results


#%% Interactive driver

if False:
    
    #%% Process images
    
    folder_name = r'g:\temp\island_conservation_camera_traps'
    output_file = r'g:\temp\ocr_results.json'
    from megadetector.utils.path_utils import insert_before_extension
    output_file = insert_before_extension(output_file)
    n_to_sample = -1
    assert os.path.isdir(folder_name)
    options_a = DatetimeExtractionOptions()
    options_b = DatetimeExtractionOptions()
    options_b.image_crop_fraction = [0.08 , 0.08]
    options_a.force_all_ocr_options = False
    options_b.force_all_ocr_options = False
    # all_options = [options_a,options_b]
    all_options = [options_a]
    filename_to_results = get_datetimes_for_folder(folder_name,output_file,
                                                   n_to_sample=n_to_sample,options=all_options)
    

    #%% Load results
    
    # output_file = r"G:\temp\ocr_results.2023.10.31.07.37.54.json"
    with open(output_file,'r') as f:
        filename_to_results = json.load(f)
    filenames = sorted(list(filename_to_results.keys()))
    print('Loaded results for {} files'.format(len(filename_to_results)))
    
    
    #%% Scrap cell
    
    fn = 'g:/camera_traps/camera_trap_images/2018.07.02/newcam/people/DSCF0273.JPG'
    include_crops = False
    options_a = DatetimeExtractionOptions()
    options_b = DatetimeExtractionOptions()
    options_b.image_crop_fraction = [0.08 , 0.08]
    image = vis_utils.open_image(fn) # noqa    
    result = try_get_datetime_from_image(fn,options=[options_a,options_b]) # noqa
    print(result)
    
    # open_file(fn)
    # rough_crops = make_rough_crops(image,options=options)
        
        
    #%% Look for OCR or parsing failures
    
    bad_tokens = ()
    
    files_with_disagreements = set()
            
    # i_fn = 0; fn = filenames[i_fn]
    for i_fn,fn in enumerate(filenames):
        
        image = fn
        results = filename_to_results[fn]
        
        if 'text_results' not in results:
            raise Exception('no results available for {} ({})'.format(i_fn,fn))
            print('Skipping {}, no results'.format(i_fn))
            continue
        
        s = ' '.join([x[0] for x in results['text_results']])
        
        known_bad = False
        for bad_token in bad_tokens:
            if bad_token in s:
                known_bad = True
        if known_bad: 
            continue
                
        extracted_datetime = results['datetime']
        
        # If we have a datetime, make sure all successful OCR results agree
        if extracted_datetime is not None:
            for config_string in results['all_extracted_datetimes']:
                if results['all_extracted_datetimes'][config_string] is not None:
                    if results['all_extracted_datetimes'][config_string] != extracted_datetime:
                        files_with_disagreements.add(fn)
        else:
            print('Falling back for {} ({})'.format(i_fn,fn))
            ocr_results = get_datetime_from_image(fn)
            extracted_datetime = ocr_results['datetime']
        
        if extracted_datetime is None:
            print('Failure at {}: {}'.format(i_fn,s))
        
        # open_file(fn)
        # get_datetime_from_image(fn)
    
    
    #%% Write results to an HTML file for testing
          
    n_to_sample = 5000
    if (n_to_sample >= 0) and (len(filename_to_results) > n_to_sample):
        filenames = sorted(list(filename_to_results.keys()))        
        import random
        random.seed(0)
        keys = random.sample(filenames,n_to_sample)
        filename_to_results = {k: filename_to_results[k] for k in keys}

    preview_dir = r'g:\temp\ocr-preview'
    os.makedirs(preview_dir,exist_ok=True)
    
    def resize_image_for_preview(fn_abs):
        fn_relative = os.path.relpath(fn_abs,folder_name)        
        resized_image = vis_utils.resize_image(fn_abs,target_width=600)
        resized_fn = os.path.join(preview_dir,fn_relative)
        os.makedirs(os.path.dirname(resized_fn),exist_ok=True)
        resized_image.save(resized_fn)
        return resized_fn
        
    # Resize images in parallel
    n_rendering_workers = 16
        
    if n_rendering_workers <= 1:
        for fn_abs in tqdm(filename_to_results.keys()):
            resize_image_for_preview(fn_abs)
    else:
        # from multiprocessing.pool import Pool as RenderingPool; worker_string = 'processes'
        from multiprocessing.pool import ThreadPool as RenderingPool; worker_string = 'threads'
        pool = RenderingPool(n_rendering_workers)
        
        print('Starting rendering pool with {} {}'.format(n_rendering_workers,worker_string))
    
        _ = list(tqdm(pool.imap(resize_image_for_preview,filename_to_results.keys()),
                      total=len(filename_to_results)))
    
    
    def make_datetime_preview_page(filenames,html_file):
        
        html_image_list = []
        html_options = write_html_image_list.write_html_image_list()
        html_options['maxFiguresPerHtmlFile'] = 2500
        html_options['defaultImageStyle'] = 'margin:0px;margin-top:5px;margin-bottom:30px;'
        
        # fn_abs = filenames[0]
        for fn_abs in filenames:
            
            fn_relative = os.path.relpath(fn_abs,folder_name)        
            # resized_fn = os.path.join(preview_dir,fn_relative)
            results_this_image = filename_to_results[fn_abs]
                
            extracted_datetime = results_this_image['datetime']
            title = 'Image: {}<br/>Extracted datetime: {}'.format(fn_relative,extracted_datetime)
            html_image_list.append({'filename':fn_relative,'title':title})
            
            # ...for each crop
        
        # ...for each image
                
        html_options['makeRelative'] = True
        write_html_image_list.write_html_image_list(html_file,
                                                    html_image_list,
                                                    html_options)
        open_file(html_file)
        return html_image_list
    
    failed_files = []
    for fn_abs in filename_to_results:
        results_this_image = filename_to_results[fn_abs]
        if results_this_image['datetime'] is None:
            failed_files.append(fn_abs)
    
    print('Found {} failures'.format(len(failed_files)))
    
    output_summary_file = os.path.join(preview_dir,'summary.html')
    html_image_list = make_datetime_preview_page(sorted(list(filename_to_results.keys())),output_summary_file)
    
    failure_summary_file = os.path.join(preview_dir,'failures.html')    
    html_image_list_failures = make_datetime_preview_page(failed_files,failure_summary_file)
    
    filenames = failed_files
    html_file = failure_summary_file

    
    #%% Other approaches to getting dates from strings
    
    # ...that didn't really work out.
    
    # pip install dateparser
    import dateparser

    # pip install datefinder
    import datefinder

    from dateparser.search import search_dates # noqa

    dateparser_settings = {'PREFER_DATES_FROM':'past','STRICT_PARSING':True}

    dateparser_result = dateparser.search.search_dates(s, settings=dateparser_settings)
    
    if dateparser_result is not None:
        assert len(dateparser_result) == 1
        extracted_datetime = dateparser_result[0][1]
    else:
        matches = datefinder.find_dates(s,strict=False)
        matches_list = [m for m in matches]
        if len(matches_list) == 1:
            extracted_datetime = matches_list[0]
        else:
            extracted_datetime = None
        
    if extracted_datetime is not None:        
        assert extracted_datetime.year <= 2023 and extracted_datetime.year >= 1990

